# Copyright 2010-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ("MaskManager",)

import warnings

from portage import os
from portage.dep import ExtendedAtomDict, match_from_list
from portage.localization import _
from portage.repository.config import allow_profile_repo_deps
from portage.util import append_repo, grabfile_package, stack_lists, writemsg
from portage.versions import _pkg_str


class MaskManager:
    def __init__(
        self,
        repositories,
        profiles,
        abs_user_config,
        user_config=True,
        strict_umatched_removal=False,
    ):
        self._punmaskdict = ExtendedAtomDict(list)
        self._pmaskdict = ExtendedAtomDict(list)
        # Preserves atoms that are eliminated by negative
        # incrementals in user_pkgmasklines.
        self._pmaskdict_raw = ExtendedAtomDict(list)

        # Read profile/package.mask from every repo.
        # Repositories inherit masks from their parent profiles and
        # are able to remove mask from them with -atoms.
        # Such a removal affects only the current repo, but not the parent.
        # Add ::repo specs to every atom to make sure atoms only affect
        # packages from the current repo.

        # Cache the repository-wide package.mask files as a particular
        # repo may be often referenced by others as the master.
        pmask_cache = {}

        def grab_pmask(loc, repo_config):
            if loc not in pmask_cache:
                path = os.path.join(loc, "profiles", "package.mask")
                pmask_cache[loc] = grabfile_package(
                    path,
                    recursive=repo_config.portage1_profiles,
                    remember_source_file=True,
                    verify_eapi=True,
                    eapi_default=repo_config.eapi,
                    allow_repo=allow_profile_repo_deps(repo_config),
                    allow_build_id=("build-id" in repo_config.profile_formats),
                )
                if repo_config.portage1_profiles_compat and os.path.isdir(path):
                    warnings.warn(
                        _(
                            "Repository '%(repo_name)s' is implicitly using "
                            "'portage-1' profile format in its profiles/package.mask, but "
                            "the repository profiles are not marked as that format.  This will break "
                            "in the future.  Please either convert the following paths "
                            "to files, or add\nprofile-formats = portage-1\nto the "
                            "repository's layout.conf.\n"
                        )
                        % dict(repo_name=repo_config.name)
                    )

            return pmask_cache[loc]

        repo_pkgmasklines = []
        for repo in repositories.repos_with_profiles():
            lines = []
            repo_lines = grab_pmask(repo.location, repo)
            removals = frozenset(
                line[0][1:] for line in repo_lines if line[0][:1] == "-"
            )
            matched_removals = set()
            for master in repo.masters:
                master_lines = grab_pmask(master.location, master)
                for line in master_lines:
                    if line[0] in removals:
                        matched_removals.add(line[0])
                # Since we don't stack masters recursively, there aren't any
                # atoms earlier in the stack to be matched by negative atoms in
                # master_lines. Also, repo_lines may contain negative atoms
                # that are intended to negate atoms from a different master
                # than the one with which we are currently stacking. Therefore,
                # we disable warn_for_unmatched_removal here (see bug #386569).
                lines.append(
                    stack_lists(
                        [master_lines, repo_lines],
                        incremental=1,
                        remember_source_file=True,
                        warn_for_unmatched_removal=False,
                    )
                )

            # It's safe to warn for unmatched removal if masters have not
            # been overridden by the user, which is guaranteed when
            # user_config is false (when called by repoman).
            if repo.masters:
                unmatched_removals = removals.difference(matched_removals)
                if unmatched_removals and not user_config:
                    source_file = os.path.join(
                        repo.location, "profiles", "package.mask"
                    )
                    unmatched_removals = list(unmatched_removals)
                    if len(unmatched_removals) > 3:
                        writemsg(
                            _("--- Unmatched removal atoms in %s: %s and %s more\n")
                            % (
                                source_file,
                                ", ".join("-" + x for x in unmatched_removals[:3]),
                                len(unmatched_removals) - 3,
                            ),
                            noiselevel=-1,
                        )
                    else:
                        writemsg(
                            _("--- Unmatched removal atom(s) in %s: %s\n")
                            % (
                                source_file,
                                ", ".join("-" + x for x in unmatched_removals),
                            ),
                            noiselevel=-1,
                        )

            else:
                lines.append(
                    stack_lists(
                        [repo_lines],
                        incremental=1,
                        remember_source_file=True,
                        warn_for_unmatched_removal=not user_config,
                        strict_warn_for_unmatched_removal=strict_umatched_removal,
                    )
                )
            repo_pkgmasklines.extend(
                append_repo(stack_lists(lines), repo.name, remember_source_file=True)
            )

        repo_pkgunmasklines = []
        for repo in repositories.repos_with_profiles():
            if not repo.portage1_profiles:
                continue
            repo_lines = grabfile_package(
                os.path.join(repo.location, "profiles", "package.unmask"),
                recursive=1,
                remember_source_file=True,
                verify_eapi=True,
                eapi_default=repo.eapi,
                allow_repo=allow_profile_repo_deps(repo),
                allow_build_id=("build-id" in repo.profile_formats),
            )
            lines = stack_lists(
                [repo_lines],
                incremental=1,
                remember_source_file=True,
                warn_for_unmatched_removal=True,
                strict_warn_for_unmatched_removal=strict_umatched_removal,
            )
            repo_pkgunmasklines.extend(
                append_repo(lines, repo.name, remember_source_file=True)
            )

        # Read package.mask from the user's profile. Stack them in the end
        # to allow profiles to override masks from their parent profiles.
        profile_pkgmasklines = []
        profile_pkgunmasklines = []
        for x in profiles:
            profile_pkgmasklines.append(
                grabfile_package(
                    os.path.join(x.location, "package.mask"),
                    recursive=x.portage1_directories,
                    remember_source_file=True,
                    verify_eapi=True,
                    eapi=x.eapi,
                    eapi_default=None,
                    allow_repo=allow_profile_repo_deps(x),
                    allow_build_id=x.allow_build_id,
                )
            )
            if x.portage1_directories:
                profile_pkgunmasklines.append(
                    grabfile_package(
                        os.path.join(x.location, "package.unmask"),
                        recursive=x.portage1_directories,
                        remember_source_file=True,
                        verify_eapi=True,
                        eapi=x.eapi,
                        eapi_default=None,
                        allow_repo=allow_profile_repo_deps(x),
                        allow_build_id=x.allow_build_id,
                    )
                )
        profile_pkgmasklines = stack_lists(
            profile_pkgmasklines,
            incremental=1,
            remember_source_file=True,
            warn_for_unmatched_removal=True,
            strict_warn_for_unmatched_removal=strict_umatched_removal,
        )
        profile_pkgunmasklines = stack_lists(
            profile_pkgunmasklines,
            incremental=1,
            remember_source_file=True,
            warn_for_unmatched_removal=True,
            strict_warn_for_unmatched_removal=strict_umatched_removal,
        )

        # Read /etc/portage/package.mask. Don't stack it to allow the user to
        # remove mask atoms from everywhere with -atoms.
        user_pkgmasklines = []
        user_pkgunmasklines = []
        if user_config:
            user_pkgmasklines = grabfile_package(
                os.path.join(abs_user_config, "package.mask"),
                recursive=1,
                allow_wildcard=True,
                allow_repo=True,
                remember_source_file=True,
                verify_eapi=False,
                allow_build_id=True,
            )
            user_pkgunmasklines = grabfile_package(
                os.path.join(abs_user_config, "package.unmask"),
                recursive=1,
                allow_wildcard=True,
                allow_repo=True,
                remember_source_file=True,
                verify_eapi=False,
                allow_build_id=True,
            )

        # Stack everything together. At this point, only user_pkgmasklines may contain -atoms.
        # Don't warn for unmatched -atoms here, since we don't do it for any other user config file.
        raw_pkgmasklines = stack_lists(
            [repo_pkgmasklines, profile_pkgmasklines],
            incremental=1,
            remember_source_file=True,
            warn_for_unmatched_removal=False,
            ignore_repo=True,
        )
        pkgmasklines = stack_lists(
            [repo_pkgmasklines, profile_pkgmasklines, user_pkgmasklines],
            incremental=1,
            remember_source_file=True,
            warn_for_unmatched_removal=False,
            ignore_repo=True,
        )
        pkgunmasklines = stack_lists(
            [repo_pkgunmasklines, profile_pkgunmasklines, user_pkgunmasklines],
            incremental=1,
            remember_source_file=True,
            warn_for_unmatched_removal=False,
            ignore_repo=True,
        )

        for x, source_file in raw_pkgmasklines:
            self._pmaskdict_raw.setdefault(x.cp, []).append(x)

        for x, source_file in pkgmasklines:
            self._pmaskdict.setdefault(x.cp, []).append(x)

        for x, source_file in pkgunmasklines:
            self._punmaskdict.setdefault(x.cp, []).append(x)

        for d in (self._pmaskdict_raw, self._pmaskdict, self._punmaskdict):
            for k, v in d.items():
                d[k] = tuple(v)

    def _getMaskAtom(self, cpv, slot, repo, unmask_atoms=None):
        """
        Take a package and return a matching package.mask atom, or None if no
        such atom exists or it has been cancelled by package.unmask.

        @param cpv: The package name
        @type cpv: String
        @param slot: The package's slot
        @type slot: String
        @param repo: The package's repository [optional]
        @type repo: String
        @param unmask_atoms: if desired pass in self._punmaskdict.get(cp)
        @type unmask_atoms: list
        @rtype: String
        @return: A matching atom string or None if one is not found.
        """

        try:
            cpv.slot
        except AttributeError:
            pkg = _pkg_str(cpv, slot=slot, repo=repo)
        else:
            pkg = cpv

        mask_atoms = self._pmaskdict.get(pkg.cp)
        if mask_atoms:
            pkg_list = [pkg]
            for x in mask_atoms:
                if not match_from_list(x, pkg_list):
                    continue
                if unmask_atoms:
                    for y in unmask_atoms:
                        if match_from_list(y, pkg_list):
                            return None
                return x
        return None

    def getMaskAtom(self, cpv, slot, repo):
        """
        Take a package and return a matching package.mask atom, or None if no
        such atom exists or it has been cancelled by package.unmask.

        @param cpv: The package name
        @type cpv: String
        @param slot: The package's slot
        @type slot: String
        @param repo: The package's repository [optional]
        @type repo: String
        @rtype: String
        @return: A matching atom string or None if one is not found.
        """

        try:
            cpv.slot
        except AttributeError:
            pkg = _pkg_str(cpv, slot=slot, repo=repo)
        else:
            pkg = cpv

        return self._getMaskAtom(pkg, slot, repo, self._punmaskdict.get(pkg.cp))

    def getRawMaskAtom(self, cpv, slot, repo):
        """
        Take a package and return a matching package.mask atom, or None if no
        such atom exists. It HAS NOT! been cancelled by any package.unmask.

        @param cpv: The package name
        @type cpv: String
        @param slot: The package's slot
        @type slot: String
        @param repo: The package's repository [optional]
        @type repo: String
        @rtype: String
        @return: A matching atom string or None if one is not found.
        """

        return self._getMaskAtom(cpv, slot, repo)



# test_isjustname.py -- Portage Unit Testing Functionality
# Copyright 2006 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.dep import isjustname


class IsJustName(TestCase):
    def testIsJustName(self):

        cats = ("", "sys-apps/", "foo/", "virtual/")
        pkgs = ("portage", "paludis", "pkgcore", "notARealPkg")
        vers = ("", "-2.0-r3", "-1.0_pre2", "-3.1b")

        for pkg in pkgs:
            for cat in cats:
                for ver in vers:
                    if len(ver):
                        self.assertFalse(
                            isjustname(cat + pkg + ver),
                            msg="isjustname(%s) is True!" % (cat + pkg + ver),
                        )
                    else:
                        self.assertTrue(
                            isjustname(cat + pkg + ver),
                            msg="isjustname(%s) is False!" % (cat + pkg + ver),
                        )



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SolveNonSlotOperatorSlotConflictsTestCase(TestCase):
    def testSolveNonSlotOperatorSlotConflicts(self):

        ebuilds = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1", "PDEPEND": "app-misc/B"},
            "app-misc/A-2": {"EAPI": "5", "SLOT": "0/2", "PDEPEND": "app-misc/B"},
            "app-misc/B-0": {"EAPI": "5", "RDEPEND": "app-misc/A:="},
        }

        installed = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1", "PDEPEND": "app-misc/B"},
            "app-misc/B-0": {"EAPI": "5", "RDEPEND": "app-misc/A:0/1="},
        }

        world = ["app-misc/A"]

        test_cases = (
            # bug 522084
            # In this case, _solve_non_slot_operator_slot_conflicts
            # removed both versions of app-misc/A from the graph, since
            # they didn't have any non-conflict parents (except for
            # @selected which matched both instances). The result was
            # a missed update.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["app-misc/A-2", "app-misc/B-0"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



#!/usr/bin/env python
# Copyright 1999-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import os
import re
import sys

here_doc_re = re.compile(rb".*\s<<[-]?(\w+)$")
func_start_re = re.compile(rb"^[-\w]+\s*\(\)\s*$")
func_end_re = re.compile(rb"^\}$")

var_assign_re = re.compile(
    rb'(^|^declare\s+-\S+\s+|^declare\s+|^export\s+)([^=\s]+)=("|\')?.*$'
)
close_quote_re = re.compile(rb'(\\"|"|\')\s*$')
readonly_re = re.compile(rb"^declare\s+-(\S*)r(\S*)\s+")
# declare without assignment
var_declare_re = re.compile(rb"^declare(\s+-\S+)?\s+([^=\s]+)\s*$")


def have_end_quote(quote, line):
    """
    Check if the line has an end quote (useful for handling multi-line
    quotes). This handles escaped double quotes that may occur at the
    end of a line. The posix spec does not allow escaping of single
    quotes inside of single quotes, so that case is not handled.
    """
    close_quote_match = close_quote_re.search(line)
    return close_quote_match is not None and close_quote_match.group(1) == quote


def filter_declare_readonly_opt(line):
    readonly_match = readonly_re.match(line)
    if readonly_match is not None:
        declare_opts = b""
        for i in (1, 2):
            group = readonly_match.group(i)
            if group is not None:
                declare_opts += group
        if declare_opts:
            line = b"declare -" + declare_opts + b" " + line[readonly_match.end() :]
        else:
            line = b"declare " + line[readonly_match.end() :]
    return line


def filter_bash_environment(pattern, file_in, file_out):
    # Filter out any instances of the \1 character from variable values
    # since this character multiplies each time that the environment
    # is saved (strange bash behavior). This can eventually result in
    # mysterious 'Argument list too long' errors from programs that have
    # huge strings of \1 characters in their environment. See bug #222091.
    here_doc_delim = None
    in_func = None
    multi_line_quote = None
    multi_line_quote_filter = None
    for line in file_in:
        if multi_line_quote is not None:
            if not multi_line_quote_filter:
                file_out.write(line.replace(b"\1", b""))
            if have_end_quote(multi_line_quote, line):
                multi_line_quote = None
                multi_line_quote_filter = None
            continue
        if here_doc_delim is None and in_func is None:
            var_assign_match = var_assign_re.match(line)
            if var_assign_match is not None:
                quote = var_assign_match.group(3)
                filter_this = pattern.match(var_assign_match.group(2)) is not None
                # Exclude the start quote when searching for the end quote,
                # to ensure that the start quote is not misidentified as the
                # end quote (happens if there is a newline immediately after
                # the start quote).
                if quote is not None and not have_end_quote(
                    quote, line[var_assign_match.end(2) + 2 :]
                ):
                    multi_line_quote = quote
                    multi_line_quote_filter = filter_this
                if not filter_this:
                    line = filter_declare_readonly_opt(line)
                    file_out.write(line.replace(b"\1", b""))
                continue
            else:
                declare_match = var_declare_re.match(line)
                if declare_match is not None:
                    # declare without assignment
                    filter_this = pattern.match(declare_match.group(2)) is not None
                    if not filter_this:
                        line = filter_declare_readonly_opt(line)
                        file_out.write(line)
                    continue

        if here_doc_delim is not None:
            if here_doc_delim.match(line):
                here_doc_delim = None
            file_out.write(line)
            continue
        here_doc = here_doc_re.match(line)
        if here_doc is not None:
            here_doc_delim = re.compile(b"^" + here_doc.group(1) + b"$")
            file_out.write(line)
            continue
        # Note: here-documents are handled before functions since otherwise
        # it would be possible for the content of a here-document to be
        # mistaken as the end of a function.
        if in_func:
            if func_end_re.match(line) is not None:
                in_func = None
            file_out.write(line)
            continue
        in_func = func_start_re.match(line)
        if in_func is not None:
            file_out.write(line)
            continue
        # This line is not recognized as part of a variable assignment,
        # function definition, or here document, so just allow it to
        # pass through.
        file_out.write(line)


if __name__ == "__main__":
    description = (
        "Filter out variable assignments for variable "
        + "names matching a given PATTERN "
        + "while leaving bash function definitions and here-documents "
        + "intact. The PATTERN is a space separated list of variable names"
        + " and it supports python regular expression syntax."
    )
    usage = "usage: %s PATTERN" % os.path.basename(sys.argv[0])
    args = sys.argv[1:]

    if "-h" in args or "--help" in args:
        sys.stdout.write(usage + "\n")
        sys.stdout.flush()
        sys.exit(os.EX_OK)

    if len(args) != 1:
        sys.stderr.write(usage + "\n")
        sys.stderr.write("Exactly one PATTERN argument required.\n")
        sys.stderr.flush()
        sys.exit(2)

    file_in = sys.stdin.buffer
    file_out = sys.stdout.buffer
    var_pattern = os.fsencode(args[0]).split()

    # Filter invalid variable names that are not supported by bash.
    var_pattern.append(rb"\d.*")
    var_pattern.append(rb".*\W.*")

    var_pattern = b"^(" + b"|".join(var_pattern) + b")$"
    filter_bash_environment(re.compile(var_pattern), file_in, file_out)
    file_out.flush()



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.versions import cpv_sort_key


class CpvSortKeyTestCase(TestCase):
    def testCpvSortKey(self):

        tests = [
            (
                ("a/b-2_alpha", "a", "b", "a/b-2", "a/a-1", "a/b-1"),
                ("a", "a/a-1", "a/b-1", "a/b-2_alpha", "a/b-2", "b"),
            ),
        ]

        for test in tests:
            self.assertEqual(tuple(sorted(test[0], key=cpv_sort_key())), test[1])



# Copyright 2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import os
import platform

import portage.process
from portage.const import BASH_BINARY
from portage.tests import TestCase

CLONE_NEWNET = 0x40000000

UNSHARE_NET_TEST_SCRIPT = """
ping -c 1 -W 1 127.0.0.1 || exit 1
ping -c 1 -W 1 10.0.0.1 || exit 1
[[ -n ${IPV6} ]] || exit 0
ping -c 1 -W 1 ::1 || exit 1
ping -c 1 -W 1 fd::1 || exit 1
"""


class UnshareNetTestCase(TestCase):
    def testUnshareNet(self):

        if platform.system() != "Linux":
            self.skipTest("not Linux")
        if portage.process.find_binary("ping") is None:
            self.skipTest("ping not found")

        errno_value = portage.process._unshare_validate(CLONE_NEWNET)
        if errno_value != 0:
            self.skipTest(
                "Unable to unshare: %s" % (errno.errorcode.get(errno_value, "?"))
            )

        env = os.environ.copy()
        env["IPV6"] = "1" if portage.process._has_ipv6() else ""
        self.assertEqual(
            portage.process.spawn(
                [BASH_BINARY, "-c", UNSHARE_NET_TEST_SCRIPT], unshare_net=True, env=env
            ),
            0,
        )



# Copyright 2012-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import errno
import re
import subprocess

from portage import os
from portage import _unicode_encode, _encodings
from portage.const import MANIFEST2_IDENTIFIERS
from portage.dep import _repo_separator
from portage.exception import InvalidDependString
from portage.localization import _
from portage.util import atomic_ofstream, grablines, shlex_split, varexpand, writemsg
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
from portage.util._async.PipeLogger import PipeLogger
from portage.util._async.PopenProcess import PopenProcess
from _emerge.CompositeTask import CompositeTask
from _emerge.PipeReader import PipeReader
from .ManifestProcess import ManifestProcess


class ManifestTask(CompositeTask):

    __slots__ = (
        "cp",
        "distdir",
        "fetchlist_dict",
        "gpg_cmd",
        "gpg_vars",
        "repo_config",
        "force_sign_key",
        "_manifest_path",
    )

    _PGP_HEADER = b"BEGIN PGP SIGNED MESSAGE"
    _manifest_line_re = re.compile(r"^(%s) " % "|".join(MANIFEST2_IDENTIFIERS))
    _gpg_key_id_re = re.compile(r"^[0-9A-F]*$")
    _gpg_key_id_lengths = (8, 16, 24, 32, 40)

    def _start(self):
        self._manifest_path = os.path.join(
            self.repo_config.location, self.cp, "Manifest"
        )

        self._start_task(
            AsyncTaskFuture(future=self.fetchlist_dict), self._start_with_fetchlist
        )

    def _start_with_fetchlist(self, fetchlist_task):
        if self._default_exit(fetchlist_task) != os.EX_OK:
            if not self.fetchlist_dict.cancelled():
                try:
                    self.fetchlist_dict.result()
                except InvalidDependString as e:
                    writemsg(
                        _("!!! %s%s%s: SRC_URI: %s\n")
                        % (self.cp, _repo_separator, self.repo_config.name, e),
                        noiselevel=-1,
                    )
            self._async_wait()
            return
        self.fetchlist_dict = self.fetchlist_dict.result()
        manifest_proc = ManifestProcess(
            cp=self.cp,
            distdir=self.distdir,
            fetchlist_dict=self.fetchlist_dict,
            repo_config=self.repo_config,
            scheduler=self.scheduler,
        )
        self._start_task(manifest_proc, self._manifest_proc_exit)

    def _manifest_proc_exit(self, manifest_proc):
        self._assert_current(manifest_proc)
        if manifest_proc.returncode not in (os.EX_OK, manifest_proc.MODIFIED):
            self.returncode = manifest_proc.returncode
            self._current_task = None
            self.wait()
            return

        modified = manifest_proc.returncode == manifest_proc.MODIFIED
        sign = self.gpg_cmd is not None

        if not modified and sign:
            sign = self._need_signature()
            if (
                not sign
                and self.force_sign_key is not None
                and os.path.exists(self._manifest_path)
            ):
                self._check_sig_key()
                return

        if not sign or not os.path.exists(self._manifest_path):
            self.returncode = os.EX_OK
            self._current_task = None
            self.wait()
            return

        self._start_gpg_proc()

    def _check_sig_key(self):
        null_fd = os.open("/dev/null", os.O_RDONLY)
        popen_proc = PopenProcess(
            proc=subprocess.Popen(
                ["gpg", "--verify", self._manifest_path],
                stdin=null_fd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            ),
            pipe_reader=PipeReader(),
        )
        os.close(null_fd)
        popen_proc.pipe_reader.input_files = {"producer": popen_proc.proc.stdout}
        self._start_task(popen_proc, self._check_sig_key_exit)

    @staticmethod
    def _parse_gpg_key(output):
        """
        Returns the first token which appears to represent a gpg key
        id, or None if there is no such token.
        """
        regex = ManifestTask._gpg_key_id_re
        lengths = ManifestTask._gpg_key_id_lengths
        for token in output.split():
            m = regex.match(token)
            if m is not None and len(m.group(0)) in lengths:
                return m.group(0)
        return None

    @staticmethod
    def _normalize_gpg_key(key_str):
        """
        Strips leading "0x" and trailing "!", and converts to uppercase
        (intended to be the same format as that in gpg --verify output).
        """
        key_str = key_str.upper()
        if key_str.startswith("0X"):
            key_str = key_str[2:]
        key_str = key_str.rstrip("!")
        return key_str

    def _check_sig_key_exit(self, proc):
        self._assert_current(proc)

        parsed_key = self._parse_gpg_key(
            proc.pipe_reader.getvalue().decode("utf_8", "replace")
        )
        if parsed_key is not None and self._normalize_gpg_key(
            parsed_key
        ) == self._normalize_gpg_key(self.force_sign_key):
            self.returncode = os.EX_OK
            self._current_task = None
            self.wait()
            return

        if self._was_cancelled():
            self.wait()
            return

        self._strip_sig(self._manifest_path)
        self._start_gpg_proc()

    @staticmethod
    def _strip_sig(manifest_path):
        """
        Strip an existing signature from a Manifest file.
        """
        line_re = ManifestTask._manifest_line_re
        lines = grablines(manifest_path)
        f = None
        try:
            f = atomic_ofstream(manifest_path)
            for line in lines:
                if line_re.match(line) is not None:
                    f.write(line)
            f.close()
            f = None
        finally:
            if f is not None:
                f.abort()

    def _start_gpg_proc(self):
        gpg_vars = self.gpg_vars
        if gpg_vars is None:
            gpg_vars = {}
        else:
            gpg_vars = gpg_vars.copy()
        gpg_vars["FILE"] = self._manifest_path
        gpg_cmd = varexpand(self.gpg_cmd, mydict=gpg_vars)
        gpg_cmd = shlex_split(gpg_cmd)
        gpg_proc = PopenProcess(
            proc=subprocess.Popen(
                gpg_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
        )
        # PipeLogger echos output and efficiently monitors for process
        # exit by listening for the stdout EOF event.
        gpg_proc.pipe_reader = PipeLogger(
            background=self.background,
            input_fd=gpg_proc.proc.stdout,
            scheduler=self.scheduler,
        )
        self._start_task(gpg_proc, self._gpg_proc_exit)

    def _gpg_proc_exit(self, gpg_proc):
        if self._default_exit(gpg_proc) != os.EX_OK:
            self.wait()
            return

        rename_args = (self._manifest_path + ".asc", self._manifest_path)
        try:
            os.rename(*rename_args)
        except OSError as e:
            writemsg("!!! rename('%s', '%s'): %s\n" % rename_args + (e,), noiselevel=-1)
            try:
                os.unlink(self._manifest_path + ".asc")
            except OSError:
                pass
            self.returncode = 1
        else:
            self.returncode = os.EX_OK

        self._current_task = None
        self.wait()

    def _need_signature(self):
        try:
            with open(
                _unicode_encode(
                    self._manifest_path, encoding=_encodings["fs"], errors="strict"
                ),
                "rb",
            ) as f:
                return self._PGP_HEADER not in f.readline()
        except IOError as e:
            if e.errno in (errno.ENOENT, errno.ESTALE):
                return False
            raise



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


from portage.dep import Atom, _repo_separator
from portage.exception import InvalidData


def create_world_atom(pkg, args_set, root_config, before_install=False):
    """Create a new atom for the world file if one does not exist.  If the
    argument atom is precise enough to identify a specific slot then a slot
    atom will be returned. Atoms that are in the system set may also be stored
    in world since a user might want to select multiple slots of a slotted
    package like gcc for example. Unslotted system packages will not be
    stored in world."""

    arg_atom = args_set.findAtomForPackage(pkg)
    if not arg_atom:
        return None
    cp = arg_atom.cp
    new_world_atom = cp
    if arg_atom.repo:
        new_world_atom += _repo_separator + arg_atom.repo
    sets = root_config.sets
    portdb = root_config.trees["porttree"].dbapi
    vardb = root_config.trees["vartree"].dbapi

    if arg_atom.repo is not None:
        repos = [arg_atom.repo]
    else:
        # Iterate over portdbapi.porttrees, since it's common to
        # tweak this attribute in order to adjust match behavior.
        repos = []
        for tree in portdb.porttrees:
            repos.append(portdb.repositories.get_name_for_location(tree))

    available_slots = set()
    for cpv in portdb.match(Atom(cp)):
        for repo in repos:
            try:
                available_slots.add(portdb._pkg_str(str(cpv), repo).slot)
            except (KeyError, InvalidData):
                pass

    slotted = len(available_slots) > 1 or (
        len(available_slots) == 1 and "0" not in available_slots
    )
    if not slotted:
        # check the vdb in case this is multislot
        available_slots = set(
            vardb._pkg_str(cpv, None).slot for cpv in vardb.match(Atom(cp))
        )
        slotted = len(available_slots) > 1 or (
            len(available_slots) == 1 and "0" not in available_slots
        )
    if slotted and arg_atom.without_repo != cp:
        # If the user gave a specific atom, store it as a
        # slot atom in the world file.
        slot_atom = pkg.slot_atom

        # For USE=multislot, there are a couple of cases to
        # handle here:
        #
        # 1) SLOT="0", but the real SLOT spontaneously changed to some
        #    unknown value, so just record an unslotted atom.
        #
        # 2) SLOT comes from an installed package and there is no
        #    matching SLOT in the ebuild repository.
        #
        # Make sure that the slot atom is available in either the
        # portdb or the vardb, since otherwise the user certainly
        # doesn't want the SLOT atom recorded in the world file
        # (case 1 above).  If it's only available in the vardb,
        # the user may be trying to prevent a USE=multislot
        # package from being removed by --depclean (case 2 above).

        mydb = portdb
        if not portdb.match(slot_atom):
            # SLOT seems to come from an installed multislot package
            mydb = vardb
        # If there is no installed package matching the SLOT atom,
        # it probably changed SLOT spontaneously due to USE=multislot,
        # so just record an unslotted atom.
        if vardb.match(slot_atom) or before_install:
            # Now verify that the argument is precise
            # enough to identify a specific slot.
            matches = mydb.match(arg_atom)
            matched_slots = set()
            if before_install:
                matched_slots.add(pkg.slot)
            if mydb is vardb:
                for cpv in matches:
                    matched_slots.add(mydb._pkg_str(cpv, None).slot)
            else:
                for cpv in matches:
                    for repo in repos:
                        try:
                            matched_slots.add(portdb._pkg_str(str(cpv), repo).slot)
                        except (KeyError, InvalidData):
                            pass

            if len(matched_slots) == 1:
                new_world_atom = slot_atom
                if arg_atom.repo:
                    new_world_atom += _repo_separator + arg_atom.repo

    if new_world_atom == sets["selected"].findAtomForPackage(pkg):
        # Both atoms would be identical, so there's nothing to add.
        return None
    if not slotted and not arg_atom.repo:
        # Don't exclude slotted atoms for system packages from world, since
        # a user might want to select multiple slots of a slotted package like
        # gcc for example.
        system_atom = sets["system"].findAtomForPackage(pkg)
        if system_atom:
            if not system_atom.cp.startswith("virtual/"):
                return None
            # System virtuals aren't safe to exclude from world since they can
            # match multiple old-style virtuals but only one of them will be
            # pulled in by update or depclean.
            providers = portdb.settings.getvirtuals().get(system_atom.cp)
            if providers and len(providers) == 1 and providers[0].cp == arg_atom.cp:
                return None
    return new_world_atom



# Copyright 2009-2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import array
import tempfile

from portage import _unicode_decode
from portage import _unicode_encode
from portage.tests import TestCase


class ArrayFromfileEofTestCase(TestCase):
    def testArrayFromfileEof(self):
        # This tests if the following python issue is fixed
        # in the currently running version of python:
        #   https://bugs.python.org/issue5334

        input_data = "an arbitrary string"
        input_bytes = _unicode_encode(input_data, encoding="utf_8", errors="strict")
        f = tempfile.TemporaryFile()
        f.write(input_bytes)

        f.seek(0)
        data = []
        eof = False
        while not eof:
            a = array.array("B")
            try:
                a.fromfile(f, len(input_bytes) + 1)
            except (EOFError, IOError):
                # python-3.0 lost data here
                eof = True

            if not a:
                eof = True
            else:
                try:
                    # Python >=3.2
                    data.append(a.tobytes())
                except AttributeError:
                    data.append(a.tostring())

        f.close()

        self.assertEqual(
            input_data,
            _unicode_decode(b"".join(data), encoding="utf_8", errors="strict"),
        )



#!/usr/bin/env python
# Copyright 2010-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import os
import sys

sys.path.insert(0, os.environ["PORTAGE_PYM_PATH"])
import portage

portage._internal_caller = True
portage._disable_legacy_globals()


def main(args):

    if args and isinstance(args[0], bytes):
        for i, x in enumerate(args):
            args[i] = portage._unicode_decode(x, errors="strict")

    # Make locks quiet since unintended locking messages displayed on
    # stdout would corrupt the intended output of this program.
    portage.locks._quiet = True
    lock_obj = portage.locks.lockfile(args[0], wantnewlockfile=True)
    sys.stdout.write("\0")
    sys.stdout.flush()
    sys.stdin.read(1)
    portage.locks.unlockfile(lock_obj)
    return portage.os.EX_OK


if __name__ == "__main__":
    rval = main(sys.argv[1:])
    sys.exit(rval)



# Copyright 2005-2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import io
import logging
import re
import subprocess

import portage
from portage import os
from portage.util import writemsg_level, shlex_split
from portage.util.futures import asyncio
from portage.output import create_color_func, EOutput

good = create_color_func("GOOD")
bad = create_color_func("BAD")
warn = create_color_func("WARN")
from portage.sync.syncbase import NewBase

try:
    from gemato.exceptions import GematoException
    import gemato.openpgp
except ImportError:
    gemato = None


class GitSync(NewBase):
    """Git sync class"""

    short_desc = "Perform sync operations on git based repositories"

    @staticmethod
    def name():
        return "GitSync"

    def __init__(self):
        NewBase.__init__(self, "git", portage.const.GIT_PACKAGE_ATOM)

    def exists(self, **kwargs):
        """Tests whether the repo actually exists"""
        return os.path.exists(os.path.join(self.repo.location, ".git"))

    def new(self, **kwargs):
        """Do the initial clone of the repository"""
        if kwargs:
            self._kwargs(kwargs)
        if not self.has_bin:
            return (1, False)
        try:
            if not os.path.exists(self.repo.location):
                os.makedirs(self.repo.location)
                self.logger(
                    self.xterm_titles, "Created new directory %s" % self.repo.location
                )
        except IOError:
            return (1, False)

        sync_uri = self.repo.sync_uri
        if sync_uri.startswith("file://"):
            sync_uri = sync_uri[7:]

        git_cmd_opts = ""
        if self.repo.module_specific_options.get("sync-git-env"):
            shlexed_env = shlex_split(self.repo.module_specific_options["sync-git-env"])
            env = dict(
                (k, v)
                for k, _, v in (assignment.partition("=") for assignment in shlexed_env)
                if k
            )
            self.spawn_kwargs["env"].update(env)

        if self.repo.module_specific_options.get("sync-git-clone-env"):
            shlexed_env = shlex_split(
                self.repo.module_specific_options["sync-git-clone-env"]
            )
            clone_env = dict(
                (k, v)
                for k, _, v in (assignment.partition("=") for assignment in shlexed_env)
                if k
            )
            self.spawn_kwargs["env"].update(clone_env)

        if self.settings.get("PORTAGE_QUIET") == "1":
            git_cmd_opts += " --quiet"
        if self.repo.clone_depth is not None:
            if self.repo.clone_depth != 0:
                git_cmd_opts += " --depth %d" % self.repo.clone_depth
        elif self.repo.sync_depth is not None:
            if self.repo.sync_depth != 0:
                git_cmd_opts += " --depth %d" % self.repo.sync_depth
        else:
            # default
            git_cmd_opts += " --depth 1"

        if self.repo.module_specific_options.get("sync-git-clone-extra-opts"):
            git_cmd_opts += (
                " %s" % self.repo.module_specific_options["sync-git-clone-extra-opts"]
            )
        git_cmd = "%s clone%s %s ." % (
            self.bin_command,
            git_cmd_opts,
            portage._shell_quote(sync_uri),
        )
        writemsg_level(git_cmd + "\n")

        exitcode = portage.process.spawn_bash(
            "cd %s ; exec %s" % (portage._shell_quote(self.repo.location), git_cmd),
            **self.spawn_kwargs,
        )
        if exitcode != os.EX_OK:
            msg = "!!! git clone error in %s" % self.repo.location
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
            return (exitcode, False)

        self.add_safe_directory()

        if not self.verify_head():
            return (1, False)

        return (os.EX_OK, True)

    def update(self):
        """Update existing git repository, and ignore the syncuri. We are
        going to trust the user and assume that the user is in the branch
        that he/she wants updated. We'll let the user manage branches with
        git directly.
        """
        if not self.has_bin:
            return (1, False)
        git_cmd_opts = ""
        quiet = self.settings.get("PORTAGE_QUIET") == "1"
        if self.repo.module_specific_options.get("sync-git-env"):
            shlexed_env = shlex_split(self.repo.module_specific_options["sync-git-env"])
            env = dict(
                (k, v)
                for k, _, v in (assignment.partition("=") for assignment in shlexed_env)
                if k
            )
            self.spawn_kwargs["env"].update(env)

        if self.repo.module_specific_options.get("sync-git-pull-env"):
            shlexed_env = shlex_split(
                self.repo.module_specific_options["sync-git-pull-env"]
            )
            pull_env = dict(
                (k, v)
                for k, _, v in (assignment.partition("=") for assignment in shlexed_env)
                if k
            )
            self.spawn_kwargs["env"].update(pull_env)

        if self.settings.get("PORTAGE_QUIET") == "1":
            git_cmd_opts += " --quiet"
        if self.repo.module_specific_options.get("sync-git-pull-extra-opts"):
            git_cmd_opts += (
                " %s" % self.repo.module_specific_options["sync-git-pull-extra-opts"]
            )

        self.add_safe_directory()

        try:
            remote_branch = portage._unicode_decode(
                subprocess.check_output(
                    [
                        self.bin_command,
                        "rev-parse",
                        "--abbrev-ref",
                        "--symbolic-full-name",
                        "@{upstream}",
                    ],
                    cwd=portage._unicode_encode(self.repo.location),
                )
            ).rstrip("\n")
        except subprocess.CalledProcessError as e:
            msg = "!!! git rev-parse error in %s" % self.repo.location
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
            return (e.returncode, False)

        shallow = self.repo.sync_depth is not None and self.repo.sync_depth != 0
        if shallow:
            git_cmd_opts += " --depth %d" % self.repo.sync_depth

            # For shallow fetch, unreachable objects may need to be pruned
            # manually, in order to prevent automatic git gc calls from
            # eventually failing (see bug 599008).
            gc_cmd = ["git", "-c", "gc.autodetach=false", "gc", "--auto"]
            if quiet:
                gc_cmd.append("--quiet")
            exitcode = portage.process.spawn(
                gc_cmd,
                cwd=portage._unicode_encode(self.repo.location),
                **self.spawn_kwargs,
            )
            if exitcode != os.EX_OK:
                msg = "!!! git gc error in %s" % self.repo.location
                self.logger(self.xterm_titles, msg)
                writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
                return (exitcode, False)

        git_cmd = "%s fetch %s%s" % (
            self.bin_command,
            remote_branch.partition("/")[0],
            git_cmd_opts,
        )

        writemsg_level(git_cmd + "\n")

        rev_cmd = [self.bin_command, "rev-list", "--max-count=1", "HEAD"]
        previous_rev = subprocess.check_output(
            rev_cmd, cwd=portage._unicode_encode(self.repo.location)
        )

        exitcode = portage.process.spawn_bash(
            "cd %s ; exec %s" % (portage._shell_quote(self.repo.location), git_cmd),
            **self.spawn_kwargs,
        )

        if exitcode != os.EX_OK:
            msg = "!!! git fetch error in %s" % self.repo.location
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
            return (exitcode, False)

        if not self.verify_head(revision="refs/remotes/%s" % remote_branch):
            return (1, False)

        if shallow:
            # Since the default merge strategy typically fails when
            # the depth is not unlimited, `git reset --merge`.
            merge_cmd = [self.bin_command, "reset", "--merge"]
        else:
            merge_cmd = [self.bin_command, "merge"]
        merge_cmd.append("refs/remotes/%s" % remote_branch)
        if quiet:
            merge_cmd.append("--quiet")
        exitcode = portage.process.spawn(
            merge_cmd,
            cwd=portage._unicode_encode(self.repo.location),
            **self.spawn_kwargs,
        )

        if exitcode != os.EX_OK:
            msg = "!!! git merge error in %s" % self.repo.location
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
            return (exitcode, False)

        current_rev = subprocess.check_output(
            rev_cmd, cwd=portage._unicode_encode(self.repo.location)
        )

        return (os.EX_OK, current_rev != previous_rev)

    def verify_head(self, revision="-1"):
        if self.repo.module_specific_options.get(
            "sync-git-verify-commit-signature", "false"
        ).lower() not in ("true", "yes"):
            return True

        if self.repo.sync_openpgp_key_path is not None and gemato is None:
            writemsg_level(
                "!!! Verifying against specified key requires gemato-14.5+ installed\n",
                level=logging.ERROR,
                noiselevel=-1,
            )
            return False

        openpgp_env = self._get_openpgp_env(self.repo.sync_openpgp_key_path)

        try:
            out = EOutput()
            env = None
            if openpgp_env is not None and self.repo.sync_openpgp_key_path is not None:
                try:
                    out.einfo("Using keys from %s" % (self.repo.sync_openpgp_key_path,))
                    with io.open(self.repo.sync_openpgp_key_path, "rb") as f:
                        openpgp_env.import_key(f)
                    self._refresh_keys(openpgp_env)
                except (GematoException, asyncio.TimeoutError) as e:
                    writemsg_level(
                        "!!! Verification impossible due to keyring problem:\n%s\n"
                        % (e,),
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    return False

                env = os.environ.copy()
                env["GNUPGHOME"] = openpgp_env.home

            rev_cmd = [self.bin_command, "log", "-n1", "--pretty=format:%G?", revision]
            try:
                status = portage._unicode_decode(
                    subprocess.check_output(
                        rev_cmd,
                        cwd=portage._unicode_encode(self.repo.location),
                        env=env,
                    )
                ).strip()
            except subprocess.CalledProcessError:
                return False

            if status == "G":  # good signature is good
                out.einfo("Trusted signature found on top commit")
                return True
            if status == "U":  # untrusted
                out.ewarn("Top commit signature is valid but not trusted")
                return True
            if status == "B":
                expl = "bad signature"
            elif status == "X":
                expl = "expired signature"
            elif status == "Y":
                expl = "expired key"
            elif status == "R":
                expl = "revoked key"
            elif status == "E":
                expl = "unable to verify signature (missing key?)"
            elif status == "N":
                expl = "no signature"
            else:
                expl = "unknown issue"
            out.eerror("No valid signature found: %s" % (expl,))
            return False
        finally:
            if openpgp_env is not None:
                openpgp_env.close()

    def retrieve_head(self, **kwargs):
        """Get information about the head commit"""
        if kwargs:
            self._kwargs(kwargs)
        if self.bin_command is None:
            # return quietly so that we don't pollute emerge --info output
            return (1, False)
        rev_cmd = [self.bin_command, "rev-list", "--max-count=1", "HEAD"]
        try:
            ret = (
                os.EX_OK,
                portage._unicode_decode(
                    subprocess.check_output(
                        rev_cmd, cwd=portage._unicode_encode(self.repo.location)
                    )
                ),
            )
        except subprocess.CalledProcessError:
            ret = (1, False)
        return ret

    def add_safe_directory(self):
        # Add safe.directory to system gitconfig if not already configured.
        # Workaround for bug #838271 and bug #838223.
        location_escaped = re.escape(self.repo.location)
        result = subprocess.run(
            [
                self.bin_command,
                "config",
                "--get",
                "safe.directory",
                f"^{location_escaped}$",
            ]
        )
        if result.returncode == 1:
            result = subprocess.run(
                [
                    self.bin_command,
                    "config",
                    "--system",
                    "--add",
                    "safe.directory",
                    self.repo.location,
                ]
            )
        return result.returncode == 0



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.util.SlotObject import SlotObject


class Task(SlotObject):
    __slots__ = ("_hash_key", "_hash_value")

    def __eq__(self, other):
        try:
            return self._hash_key == other._hash_key
        except AttributeError:
            # depgraph._pkg() generates _hash_key
            # for lookups here, so handle that
            return self._hash_key == other

    def __ne__(self, other):
        try:
            return self._hash_key != other._hash_key
        except AttributeError:
            return True

    def __hash__(self):
        return self._hash_value

    def __len__(self):
        return len(self._hash_key)

    def __getitem__(self, key):
        return self._hash_key[key]

    def __iter__(self):
        return iter(self._hash_key)

    def __contains__(self, key):
        return key in self._hash_key

    def __str__(self):
        """
        Emulate tuple.__repr__, but don't show 'foo' as u'foo' for unicode
        strings.
        """
        return "(%s)" % ", ".join(("'%s'" % x for x in self._hash_key))

    def __repr__(self):
        if self._hash_key is None:
            # triggered by python-trace
            return SlotObject.__repr__(self)
        return "<%s (%s)>" % (
            self.__class__.__name__,
            ", ".join(("'%s'" % x for x in self._hash_key)),
        )



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SimpleDepcleanTestCase(TestCase):
    def testSimpleDepclean(self):
        ebuilds = {
            "dev-libs/A-1": {},
            "dev-libs/B-1": {},
        }
        installed = {
            "dev-libs/A-1": {},
            "dev-libs/B-1": {},
        }

        world = ("dev-libs/A",)

        test_cases = (
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=["dev-libs/B-1"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()


class DepcleanWithDepsTestCase(TestCase):
    def testDepcleanWithDeps(self):
        ebuilds = {
            "dev-libs/A-1": {"RDEPEND": "dev-libs/C"},
            "dev-libs/B-1": {"RDEPEND": "dev-libs/D"},
            "dev-libs/C-1": {},
            "dev-libs/D-1": {"RDEPEND": "dev-libs/E"},
            "dev-libs/E-1": {"RDEPEND": "dev-libs/F"},
            "dev-libs/F-1": {},
        }
        installed = {
            "dev-libs/A-1": {"RDEPEND": "dev-libs/C"},
            "dev-libs/B-1": {"RDEPEND": "dev-libs/D"},
            "dev-libs/C-1": {},
            "dev-libs/D-1": {"RDEPEND": "dev-libs/E"},
            "dev-libs/E-1": {"RDEPEND": "dev-libs/F"},
            "dev-libs/F-1": {},
        }

        world = ("dev-libs/A",)

        test_cases = (
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=[
                    "dev-libs/B-1",
                    "dev-libs/D-1",
                    "dev-libs/E-1",
                    "dev-libs/F-1",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()


class DepcleanWithInstalledMaskedTestCase(TestCase):
    def testDepcleanWithInstalledMasked(self):
        """
        Test case for bug 332719.
        emerge --declean ignores that B is masked by license and removes C.
        The next emerge -uDN world doesn't take B and installs C again.
        """
        ebuilds = {
            "dev-libs/A-1": {"RDEPEND": "|| ( dev-libs/B dev-libs/C )"},
            "dev-libs/B-1": {"LICENSE": "TEST", "KEYWORDS": "x86"},
            "dev-libs/C-1": {"KEYWORDS": "x86"},
        }
        installed = {
            "dev-libs/A-1": {"RDEPEND": "|| ( dev-libs/B dev-libs/C )"},
            "dev-libs/B-1": {"LICENSE": "TEST", "KEYWORDS": "x86"},
            "dev-libs/C-1": {"KEYWORDS": "x86"},
        }

        world = ("dev-libs/A",)

        test_cases = (
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                # cleanlist=["dev-libs/C-1"]),
                cleanlist=["dev-libs/B-1"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()


class DepcleanInstalledKeywordMaskedSlotTestCase(TestCase):
    def testDepcleanInstalledKeywordMaskedSlot(self):
        """
        Verify that depclean removes newer slot
        masked by KEYWORDS (see bug #350285).
        """
        ebuilds = {
            "dev-libs/A-1": {"RDEPEND": "|| ( =dev-libs/B-2.7* =dev-libs/B-2.6* )"},
            "dev-libs/B-2.6": {"SLOT": "2.6", "KEYWORDS": "x86"},
            "dev-libs/B-2.7": {"SLOT": "2.7", "KEYWORDS": "~x86"},
        }
        installed = {
            "dev-libs/A-1": {
                "EAPI": "3",
                "RDEPEND": "|| ( dev-libs/B:2.7 dev-libs/B:2.6 )",
            },
            "dev-libs/B-2.6": {"SLOT": "2.6", "KEYWORDS": "x86"},
            "dev-libs/B-2.7": {"SLOT": "2.7", "KEYWORDS": "~x86"},
        }

        world = ("dev-libs/A",)

        test_cases = (
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=["dev-libs/B-2.7"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()


class DepcleanWithExcludeTestCase(TestCase):
    def testDepcleanWithExclude(self):

        installed = {
            "dev-libs/A-1": {},
            "dev-libs/B-1": {"RDEPEND": "dev-libs/A"},
        }

        # depclean asserts non-empty @world set
        world = ["non-empty/world-set"]

        test_cases = (
            # Without --exclude.
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=["dev-libs/B-1", "dev-libs/A-1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"], options={"--depclean": True}, success=True, cleanlist=[]
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                options={"--depclean": True},
                success=True,
                cleanlist=["dev-libs/B-1"],
            ),
            # With --exclude
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True, "--exclude": ["dev-libs/A"]},
                success=True,
                cleanlist=["dev-libs/B-1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                options={"--depclean": True, "--exclude": ["dev-libs/B"]},
                success=True,
                cleanlist=[],
            ),
        )

        playground = ResolverPlayground(installed=installed, world=world)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()


class DepcleanWithExcludeAndSlotsTestCase(TestCase):
    def testDepcleanWithExcludeAndSlots(self):

        installed = {
            "dev-libs/Z-1": {"SLOT": 1},
            "dev-libs/Z-2": {"SLOT": 2},
            "dev-libs/Y-1": {"RDEPEND": "=dev-libs/Z-1", "SLOT": 1},
            "dev-libs/Y-2": {"RDEPEND": "=dev-libs/Z-2", "SLOT": 2},
        }

        world = ["dev-libs/Y"]

        test_cases = (
            # Without --exclude.
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=["dev-libs/Y-1", "dev-libs/Z-1"],
            ),
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True, "--exclude": ["dev-libs/Z"]},
                success=True,
                cleanlist=["dev-libs/Y-1"],
            ),
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True, "--exclude": ["dev-libs/Y"]},
                success=True,
                cleanlist=[],
            ),
        )

        playground = ResolverPlayground(installed=installed, world=world)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()


class DepcleanAndWildcardsTestCase(TestCase):
    def testDepcleanAndWildcards(self):

        installed = {
            "dev-libs/A-1": {"RDEPEND": "dev-libs/B"},
            "dev-libs/B-1": {},
        }

        # depclean asserts non-empty @world set
        world = ["non-empty/world-set"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["*/*"],
                options={"--depclean": True},
                success=True,
                cleanlist=["dev-libs/A-1", "dev-libs/B-1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/*"],
                options={"--depclean": True},
                success=True,
                cleanlist=["dev-libs/A-1", "dev-libs/B-1"],
            ),
            ResolverPlaygroundTestCase(
                ["*/A"],
                options={"--depclean": True},
                success=True,
                cleanlist=["dev-libs/A-1"],
            ),
            ResolverPlaygroundTestCase(
                ["*/B"], options={"--depclean": True}, success=True, cleanlist=[]
            ),
        )

        playground = ResolverPlayground(installed=installed, world=world)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2005-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage
from portage import os
from portage.util import shlex_split, varexpand

# default clean command from make.globals
## PORTAGE_LOGDIR_CLEAN = 'find "${PORTAGE_LOGDIR}" -type f ! -name "summary.log*" -mtime +7 -delete'

ERROR_MESSAGES = {
    78: "PORTAGE_LOGDIR variable not set or PORTAGE_LOGDIR not a directory.",
    127: "PORTAGE_LOGDIR_CLEAN command not found.",
}


class CleanLogs:

    short_desc = "Clean PORTAGE_LOGDIR logs"

    @staticmethod
    def name():
        return "logs"

    def can_progressbar(self, func):
        return False

    def check(self, **kwargs):
        options = kwargs.get("options", None)
        if options:
            options["pretend"] = True
        return self.clean(**kwargs)

    def clean(self, **kwargs):
        """Log directory cleaning function

        @param **kwargs: optional dictionary of values used in this function are:
                settings: portage settings instance: defaults to portage.settings
                        "PORTAGE_LOGDIR": directory to clean
                        "PORTAGE_LOGDIR_CLEAN": command for cleaning the logs.
                options: dict:
                        'NUM': int: number of days
                        'pretend': boolean
        """
        num_of_days = None
        pretend = False

        # convoluted, I know, but portage.settings does not exist in
        # kwargs.get() when called from _emerge.main.clean_logs()
        settings = kwargs.get("settings", getattr(portage, "settings", {}))

        options = kwargs.get("options", None)
        if options:
            num_of_days = options.get("NUM", None)
            pretend = options.get("pretend", False)

        clean_cmd = settings.get("PORTAGE_LOGDIR_CLEAN")
        if clean_cmd:
            clean_cmd = shlex_split(clean_cmd)
            if "-mtime" in clean_cmd and num_of_days is not None:
                if num_of_days == 0:
                    i = clean_cmd.index("-mtime")
                    clean_cmd.remove("-mtime")
                    clean_cmd.pop(i)
                else:
                    clean_cmd[clean_cmd.index("-mtime") + 1] = "+%s" % str(num_of_days)
            if pretend:
                if "-delete" in clean_cmd:
                    clean_cmd.remove("-delete")

        if not clean_cmd:
            return (True, None)
        rval = self._clean_logs(clean_cmd, settings)
        errors = self._convert_errors(rval)
        if errors:
            return (False, errors)
        return (True, None)

    @staticmethod
    def _clean_logs(clean_cmd, settings):
        logdir = settings.get("PORTAGE_LOGDIR")
        if logdir is None or not os.path.isdir(logdir):
            return 78

        variables = {"PORTAGE_LOGDIR": logdir}
        cmd = [varexpand(x, mydict=variables) for x in clean_cmd]

        try:
            rval = portage.process.spawn(cmd, env=os.environ)
        except portage.exception.CommandNotFound:
            rval = 127
        return rval

    @staticmethod
    def _convert_errors(rval):
        msg = []
        if rval != os.EX_OK:
            if rval in ERROR_MESSAGES:
                msg.append(ERROR_MESSAGES[rval])
            else:
                msg.append("PORTAGE_LOGDIR_CLEAN command returned %s" % rval)
            msg.append(
                "See the make.conf(5) man page for "
                "PORTAGE_LOGDIR_CLEAN usage instructions."
            )
        return msg



# Copyright 2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class BacktrackMissedUpdateTestCase(TestCase):
    def testBacktrackMissedUpdateTestCase(self):

        ebuilds = {
            "dev-lang/python-2.7.18-r2": {
                "EAPI": "7",
                "SLOT": "2.7",
            },
            "dev-python/pypy3-7.3.2_rc2_p37-r1": {
                "EAPI": "7",
                "SLOT": "0/pypy37-pp73",
            },
            "dev-python/pypy3-7.3.1-r3": {
                "EAPI": "7",
                "SLOT": "0/pypy36-pp73",
            },
            "dev-python/setuptools-50.3.0": {
                "EAPI": "7",
                "IUSE": "python_targets_pypy3",
                "RDEPEND": "python_targets_pypy3? ( dev-python/pypy3:= )",
            },
            "dev-python/setuptools-50.2.0": {
                "EAPI": "7",
                "IUSE": "python_targets_pypy3",
                "RDEPEND": "python_targets_pypy3? ( dev-python/pypy3:= )",
            },
            "dev-python/setuptools-50.1.0": {
                "EAPI": "7",
                "IUSE": "python_targets_pypy3",
                "RDEPEND": "python_targets_pypy3? ( dev-python/pypy3:= )",
            },
            "dev-python/setuptools-49.6.0": {
                "EAPI": "7",
                "IUSE": "python_targets_pypy3",
                "RDEPEND": "python_targets_pypy3? ( dev-python/pypy3:= )",
            },
            "dev-python/setuptools-46.4.0-r2": {
                "EAPI": "7",
                "IUSE": "+python_targets_pypy3 +python_targets_python2_7",
                "RDEPEND": "python_targets_pypy3? ( dev-python/pypy3:= )",
            },
            "dev-lang/python-2.7.18-r2": {
                "EAPI": "7",
                "IUSE": "+python_targets_pypy3 +python_targets_python2_7",
                "RDEPEND": "python_targets_pypy3? ( dev-python/pypy3:= )",
            },
            "dev-vcs/mercurial-5.5.1": {
                "EAPI": "7",
                "IUSE": "+python_targets_pypy3 +python_targets_python2_7",
                "RDEPEND": "dev-python/setuptools[python_targets_pypy3?,python_targets_python2_7?] python_targets_python2_7? ( dev-lang/python:2.7 ) python_targets_pypy3? ( dev-python/pypy3:= )",
            },
        }

        installed = {
            "dev-lang/python-2.7.18-r2": {
                "EAPI": "7",
                "SLOT": "2.7",
            },
            "dev-python/pypy3-7.3.1-r3": {
                "EAPI": "7",
                "SLOT": "0/pypy36-pp73",
            },
            "dev-python/setuptools-46.4.0-r2": {
                "EAPI": "7",
                "IUSE": "+python_targets_pypy3 +python_targets_python2_7",
                "USE": "python_targets_pypy3 python_targets_python2_7",
                "RDEPEND": "dev-python/pypy3:0/pypy36-pp73=",
            },
            "dev-vcs/mercurial-5.5.1": {
                "EAPI": "7",
                "IUSE": "+python_targets_pypy3 +python_targets_python2_7",
                "USE": "python_targets_pypy3 python_targets_python2_7",
                "RDEPEND": "dev-python/setuptools[python_targets_pypy3,python_targets_python2_7] dev-python/pypy3:0/pypy36-pp73=",
            },
        }

        world = ["dev-vcs/mercurial"]

        test_cases = (
            # Bug 743115: missed updates trigger excessive backtracking
            ResolverPlaygroundTestCase(
                [">=dev-python/pypy3-7.3.2_rc", "@world"],
                options={"--update": True, "--deep": True, "--backtrack": 4},
                success=True,
                mergelist=[
                    "dev-python/pypy3-7.3.2_rc2_p37-r1",
                    "dev-python/setuptools-46.4.0-r2",
                    "dev-vcs/mercurial-5.5.1",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 1999-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.EbuildPhase import EbuildPhase
from _emerge.TaskSequence import TaskSequence
from _emerge.CompositeTask import CompositeTask
import portage
from portage import os
from portage.eapi import (
    eapi_has_src_prepare_and_src_configure,
    eapi_exports_replace_vars,
)


class EbuildExecuter(CompositeTask):

    __slots__ = ("pkg", "settings")

    _phases = ("prepare", "configure", "compile", "test", "install")

    def _start(self):
        pkg = self.pkg
        scheduler = self.scheduler
        settings = self.settings
        cleanup = 0
        portage.prepare_build_dirs(pkg.root, settings, cleanup)

        if eapi_exports_replace_vars(settings["EAPI"]):
            vardb = pkg.root_config.trees["vartree"].dbapi
            settings["REPLACING_VERSIONS"] = " ".join(
                set(
                    portage.versions.cpv_getversion(match)
                    for match in vardb.match(pkg.slot_atom) + vardb.match("=" + pkg.cpv)
                )
            )

        setup_phase = EbuildPhase(
            background=self.background,
            phase="setup",
            scheduler=scheduler,
            settings=settings,
        )

        setup_phase.addExitListener(self._setup_exit)
        self._task_queued(setup_phase)
        self.scheduler.scheduleSetup(setup_phase)

    def _setup_exit(self, setup_phase):

        if self._default_exit(setup_phase) != os.EX_OK:
            self.wait()
            return

        unpack_phase = EbuildPhase(
            background=self.background,
            phase="unpack",
            scheduler=self.scheduler,
            settings=self.settings,
        )

        if "live" in self.settings.get("PROPERTIES", "").split():
            # Serialize $DISTDIR access for live ebuilds since
            # otherwise they can interfere with eachother.

            unpack_phase.addExitListener(self._unpack_exit)
            self._task_queued(unpack_phase)
            self.scheduler.scheduleUnpack(unpack_phase)

        else:
            self._start_task(unpack_phase, self._unpack_exit)

    def _unpack_exit(self, unpack_phase):

        if self._default_exit(unpack_phase) != os.EX_OK:
            self.wait()
            return

        ebuild_phases = TaskSequence(scheduler=self.scheduler)

        pkg = self.pkg
        phases = self._phases
        eapi = pkg.eapi
        if not eapi_has_src_prepare_and_src_configure(eapi):
            # skip src_prepare and src_configure
            phases = phases[2:]

        for phase in phases:
            ebuild_phases.add(
                EbuildPhase(
                    background=self.background,
                    phase=phase,
                    scheduler=self.scheduler,
                    settings=self.settings,
                )
            )

        self._start_task(ebuild_phases, self._default_final_exit)



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.SubProcess import SubProcess
import sys
from portage.cache.mappings import slot_dict_class
import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.package.ebuild._metadata_invalid:eapi_invalid",
)
from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode

import fcntl
import io


class EbuildMetadataPhase(SubProcess):

    """
    Asynchronous interface for the ebuild "depend" phase which is
    used to extract metadata from the ebuild.
    """

    __slots__ = (
        "cpv",
        "eapi_supported",
        "ebuild_hash",
        "fd_pipes",
        "metadata",
        "portdb",
        "repo_path",
        "settings",
        "write_auxdb",
    ) + (
        "_eapi",
        "_eapi_lineno",
        "_raw_metadata",
    )

    _file_names = ("ebuild",)
    _files_dict = slot_dict_class(_file_names, prefix="")

    def _start(self):
        ebuild_path = self.ebuild_hash.location

        with io.open(
            _unicode_encode(ebuild_path, encoding=_encodings["fs"], errors="strict"),
            mode="r",
            encoding=_encodings["repo.content"],
            errors="replace",
        ) as f:
            self._eapi, self._eapi_lineno = portage._parse_eapi_ebuild_head(f)

        parsed_eapi = self._eapi
        if parsed_eapi is None:
            parsed_eapi = "0"

        if not parsed_eapi:
            # An empty EAPI setting is invalid.
            self._eapi_invalid(None)
            self.returncode = 1
            self._async_wait()
            return

        self.eapi_supported = portage.eapi_is_supported(parsed_eapi)
        if not self.eapi_supported:
            self.metadata = {"EAPI": parsed_eapi}
            self.returncode = os.EX_OK
            self._async_wait()
            return

        settings = self.settings
        settings.setcpv(self.cpv)
        settings.configdict["pkg"]["EAPI"] = parsed_eapi

        debug = settings.get("PORTAGE_DEBUG") == "1"
        master_fd = None
        slave_fd = None
        fd_pipes = None
        if self.fd_pipes is not None:
            fd_pipes = self.fd_pipes.copy()
        else:
            fd_pipes = {}

        null_input = open("/dev/null", "rb")
        fd_pipes.setdefault(0, null_input.fileno())
        fd_pipes.setdefault(1, sys.__stdout__.fileno())
        fd_pipes.setdefault(2, sys.__stderr__.fileno())

        # flush any pending output
        stdout_filenos = (sys.__stdout__.fileno(), sys.__stderr__.fileno())
        for fd in fd_pipes.values():
            if fd in stdout_filenos:
                sys.__stdout__.flush()
                sys.__stderr__.flush()
                break

        self._files = self._files_dict()
        files = self._files

        master_fd, slave_fd = os.pipe()

        fcntl.fcntl(
            master_fd,
            fcntl.F_SETFL,
            fcntl.fcntl(master_fd, fcntl.F_GETFL) | os.O_NONBLOCK,
        )

        fd_pipes[slave_fd] = slave_fd
        settings["PORTAGE_PIPE_FD"] = str(slave_fd)

        self._raw_metadata = []
        files.ebuild = master_fd
        self.scheduler.add_reader(files.ebuild, self._output_handler)
        self._registered = True

        retval = portage.doebuild(
            ebuild_path,
            "depend",
            settings=settings,
            debug=debug,
            mydbapi=self.portdb,
            tree="porttree",
            fd_pipes=fd_pipes,
            returnpid=True,
        )
        settings.pop("PORTAGE_PIPE_FD", None)

        os.close(slave_fd)
        null_input.close()

        if isinstance(retval, int):
            # doebuild failed before spawning
            self.returncode = retval
            self._async_wait()
            return

        self.pid = retval[0]

    def _output_handler(self):
        while True:
            buf = self._read_buf(self._files.ebuild)
            if buf is None:
                break  # EAGAIN
            elif buf:
                self._raw_metadata.append(buf)
            else:  # EIO/POLLHUP
                if self.pid is None:
                    self._unregister()
                    self._async_wait()
                else:
                    self._async_waitpid()
                break

    def _unregister(self):
        if self._files is not None:
            self.scheduler.remove_reader(self._files.ebuild)
        SubProcess._unregister(self)

    def _async_waitpid_cb(self, *args, **kwargs):
        """
        Override _async_waitpid_cb to perform cleanup that is
        not necessarily idempotent.
        """
        SubProcess._async_waitpid_cb(self, *args, **kwargs)
        # self._raw_metadata is None when _start returns
        # early due to an unsupported EAPI
        if self.returncode == os.EX_OK and self._raw_metadata is not None:
            metadata_lines = _unicode_decode(
                b"".join(self._raw_metadata),
                encoding=_encodings["repo.content"],
                errors="replace",
            ).splitlines()
            metadata = {}
            metadata_valid = True
            for l in metadata_lines:
                if "=" not in l:
                    metadata_valid = False
                    break
                key, value = l.split("=", 1)
                metadata[key] = value

            if metadata_valid:
                parsed_eapi = self._eapi
                if parsed_eapi is None:
                    parsed_eapi = "0"
                self.eapi_supported = portage.eapi_is_supported(metadata["EAPI"])
                if (not metadata["EAPI"] or self.eapi_supported) and metadata[
                    "EAPI"
                ] != parsed_eapi:
                    self._eapi_invalid(metadata)
                    metadata_valid = False

            if metadata_valid:
                # Since we're supposed to be able to efficiently obtain the
                # EAPI from _parse_eapi_ebuild_head, we don't write cache
                # entries for unsupported EAPIs.
                if self.eapi_supported:

                    if metadata.get("INHERITED", False):
                        metadata[
                            "_eclasses_"
                        ] = self.portdb.repositories.get_repo_for_location(
                            self.repo_path
                        ).eclass_db.get_eclass_data(
                            metadata["INHERITED"].split()
                        )
                    else:
                        metadata["_eclasses_"] = {}
                    metadata.pop("INHERITED", None)

                    # If called by egencache, this cache write is
                    # undesirable when metadata-transfer is disabled.
                    if self.write_auxdb is not False:
                        self.portdb._write_cache(
                            self.cpv, self.repo_path, metadata, self.ebuild_hash
                        )
                else:
                    metadata = {"EAPI": metadata["EAPI"]}
                self.metadata = metadata
            else:
                self.returncode = 1

    def _eapi_invalid(self, metadata):
        repo_name = self.portdb.getRepositoryName(self.repo_path)
        if metadata is not None:
            eapi_var = metadata["EAPI"]
        else:
            eapi_var = None
        eapi_invalid(
            self,
            self.cpv,
            repo_name,
            self.settings,
            eapi_var,
            self._eapi,
            self._eapi_lineno,
        )



# Copyright 2020-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import sys

import portage
from portage import os
from portage.tests import TestCase
from portage.util._async.AsyncFunction import AsyncFunction
from portage.util.futures import asyncio
from portage.util.futures._asyncio.streams import _writer
from portage.util.futures.unix_events import _set_nonblocking


class AsyncFunctionTestCase(TestCase):
    @staticmethod
    def _read_from_stdin(pw):
        os.close(pw)
        return "".join(sys.stdin)

    async def _testAsyncFunctionStdin(self, loop):
        test_string = "1\n2\n3\n"
        pr, pw = os.pipe()
        fd_pipes = {0: pr}
        reader = AsyncFunction(
            scheduler=loop, fd_pipes=fd_pipes, target=self._read_from_stdin, args=(pw,)
        )
        reader.start()
        os.close(pr)
        _set_nonblocking(pw)
        with open(pw, mode="wb", buffering=0) as pipe_write:
            await _writer(pipe_write, test_string.encode("utf_8"))
        self.assertEqual((await reader.async_wait()), os.EX_OK)
        self.assertEqual(reader.result, test_string)

    def testAsyncFunctionStdin(self):
        loop = asyncio._wrap_loop()
        loop.run_until_complete(self._testAsyncFunctionStdin(loop=loop))

    def _test_getpid_fork(self):
        """
        Verify that portage.getpid() cache is updated in a forked child process.
        """
        loop = asyncio._wrap_loop()
        proc = AsyncFunction(scheduler=loop, target=portage.getpid)
        proc.start()
        proc.wait()
        self.assertEqual(proc.pid, proc.result)

    def test_getpid_fork(self):
        self._test_getpid_fork()

    def test_getpid_double_fork(self):
        """
        Verify that portage.getpid() cache is updated correctly after
        two forks.
        """
        loop = asyncio._wrap_loop()
        proc = AsyncFunction(scheduler=loop, target=self._test_getpid_fork)
        proc.start()
        self.assertEqual(proc.wait(), 0)



# Copyright 2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from __future__ import print_function
import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SlotChangeWithoutRevBumpTestCase(TestCase):
    def testSlotChangeWithoutRevBump(self):

        ebuilds = {
            "app-arch/libarchive-3.1.1": {"EAPI": "5", "SLOT": "0/13"},
            "app-arch/libarchive-3.0.4-r1": {"EAPI": "5", "SLOT": "0"},
            "kde-base/ark-4.10.0": {
                "EAPI": "5",
                "DEPEND": "app-arch/libarchive:=",
                "RDEPEND": "app-arch/libarchive:=",
            },
        }

        binpkgs = {
            "app-arch/libarchive-3.1.1": {"EAPI": "5", "SLOT": "0"},
        }

        installed = {
            "app-arch/libarchive-3.1.1": {"EAPI": "5", "SLOT": "0"},
            "kde-base/ark-4.10.0": {
                "EAPI": "5",
                "DEPEND": "app-arch/libarchive:0/0=",
                "RDEPEND": "app-arch/libarchive:0/0=",
            },
        }

        world = ["kde-base/ark"]

        test_cases = (
            # Demonstrate bug #456208, where a sub-slot change
            # without revbump needs to trigger a rebuild.
            ResolverPlaygroundTestCase(
                ["kde-base/ark"],
                options={"--oneshot": True, "--usepkg": True},
                success=True,
                mergelist=["app-arch/libarchive-3.1.1", "kde-base/ark-4.10.0"],
            ),
            ResolverPlaygroundTestCase(
                ["app-arch/libarchive"],
                options={"--noreplace": True, "--usepkg": True},
                success=True,
                mergelist=[],
            ),
            ResolverPlaygroundTestCase(
                ["app-arch/libarchive"],
                options={"--usepkg": True},
                success=True,
                mergelist=["[binary]app-arch/libarchive-3.1.1"],
            ),
            # Test --changed-slot
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--changed-slot": True,
                    "--usepkg": True,
                    "--update": True,
                    "--deep": True,
                },
                success=True,
                mergelist=["app-arch/libarchive-3.1.1", "kde-base/ark-4.10.0"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import bisect


class MultiIterGroupBy:
    """
    This class functions similarly to the itertools.groupby function,
    except that it takes multiple source iterators as input. The source
    iterators must yield objects in sorted order. A group is yielded as
    soon as the progress of all iterators reaches a state which
    guarantees that there can not be any remaining (unseen) elements of
    the group. This is useful for incremental display of grouped search
    results.
    """

    def __init__(self, iterators, key=None):
        self._iterators = iterators
        self._key = key

    def __iter__(self):

        trackers = []
        for iterator in self._iterators:
            trackers.append(_IteratorTracker(iterator))

        key_map = {}
        key_list = []
        eof = []
        key_getter = self._key
        if key_getter is None:
            key_getter = lambda x: x
        min_progress = None

        while trackers:

            for tracker in trackers:

                if tracker.current is not None and tracker.current != min_progress:
                    # The trackers are sorted by progress, so the
                    # remaining trackers are guaranteed to have
                    # sufficient progress.
                    break

                # In order to avoid over-buffering (waste of memory),
                # only grab a single entry.
                try:
                    entry = next(tracker.iterator)
                except StopIteration:
                    eof.append(tracker)
                else:
                    tracker.current = key_getter(entry)
                    key_group = key_map.get(tracker.current)
                    if key_group is None:
                        key_group = []
                        key_map[tracker.current] = key_group
                        bisect.insort(key_list, tracker.current)
                    key_group.append(entry)

            if eof:
                for tracker in eof:
                    trackers.remove(tracker)
                del eof[:]

            if trackers:
                trackers.sort()
                min_progress = trackers[0].current
                # yield if key <= min_progress
                i = bisect.bisect_right(key_list, min_progress)
                yield_these = key_list[:i]
                del key_list[:i]
            else:
                yield_these = key_list
                key_list = []

            if yield_these:
                for k in yield_these:
                    yield key_map.pop(k)


class _IteratorTracker:

    __slots__ = ("current", "iterator")

    def __init__(self, iterator):

        self.iterator = iterator
        self.current = None

    def __lt__(self, other):
        if self.current is None:
            return other.current is not None
        return other.current is not None and self.current < other.current



# Copyright 2013-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os as _os

from portage import _encodings, _unicode_encode
from portage.util import apply_stat_permissions
from portage.util.file_copy import copyfile
from portage.util.futures import asyncio
from portage.util.futures.executor.fork import ForkExecutor
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture


class FileCopier(AsyncTaskFuture):
    """
    Asynchronously copy a file.
    """

    __slots__ = ("src_path", "dest_path")

    def _start(self):
        self.future = asyncio.ensure_future(
            self.scheduler.run_in_executor(ForkExecutor(loop=self.scheduler), self._run)
        )
        super(FileCopier, self)._start()

    def _run(self):
        src_path = _unicode_encode(
            self.src_path, encoding=_encodings["fs"], errors="strict"
        )
        dest_path = _unicode_encode(
            self.dest_path, encoding=_encodings["fs"], errors="strict"
        )
        copyfile(src_path, dest_path)
        apply_stat_permissions(dest_path, _os.stat(src_path))



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import sys
import time

from portage.output import colorize


def countdown(secs=5, doing="Starting"):
    if secs:
        print(
            ">>> Waiting %s seconds before starting...\n"
            ">>> (Control-C to abort)...\n"
            "%s in:" % (secs, doing),
            end="",
        )
        for sec in range(secs, 0, -1):
            sys.stdout.write(colorize("UNMERGE_WARN", " %i" % sec))
            sys.stdout.flush()
            time.sleep(1)
        print()



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import os
import shutil
import tempfile

try:
    from portage.util.file_copy.reflink_linux import file_copy as _file_copy
except ImportError:
    _file_copy = None


def _optimized_copyfile(src, dst):
    """
    Copy the contents (no metadata) of the file named src to a file
    named dst.

    If possible, copying is done within the kernel, and uses
    "copy acceleration" techniques (such as reflinks). This also
    supports sparse files.

    @param src: path of source file
    @type src: str
    @param dst: path of destination file
    @type dst: str
    """
    with open(src, "rb", buffering=0) as src_file, open(
        dst, "wb", buffering=0
    ) as dst_file:
        _file_copy(src_file.fileno(), dst_file.fileno())


if _file_copy is None:
    copyfile = shutil.copyfile
else:
    copyfile = _optimized_copyfile



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage import os

getloadavg = getattr(os, "getloadavg", None)
if getloadavg is None:

    def getloadavg():
        """
        Uses /proc/loadavg to emulate os.getloadavg().
        Raises OSError if the load average was unobtainable.
        """
        try:
            with open("/proc/loadavg") as f:
                loadavg_str = f.readline()
        except IOError:
            # getloadavg() is only supposed to raise OSError, so convert
            raise OSError("unknown")
        loadavg_split = loadavg_str.split()
        if len(loadavg_split) < 3:
            raise OSError("unknown")
        loadavg_floats = []
        for i in range(3):
            try:
                loadavg_floats.append(float(loadavg_split[i]))
            except ValueError:
                raise OSError("unknown")
        return tuple(loadavg_floats)



# test_stackLists.py -- Portage Unit Testing Functionality
# Copyright 2006-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.util import stack_lists


class StackListsTestCase(TestCase):
    def testStackLists(self):

        tests = [
            ([["a", "b", "c"], ["d", "e", "f"]], ["a", "c", "b", "e", "d", "f"], False),
            ([["a", "x"], ["b", "x"]], ["a", "x", "b"], False),
            ([["a", "b", "c"], ["-*"]], [], True),
            ([["a"], ["-a"]], [], True),
        ]

        for test in tests:
            result = stack_lists(test[0], test[2])
            self.assertEqual(set(result), set(test[1]))



# Copyright 2011-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage
from portage.dep import Atom, _get_useflag_re
from portage.eapi import _get_eapi_attrs


def expand_new_virt(vardb, atom):
    """
    Iterate over the recursively expanded RDEPEND atoms of
    a new-style virtual. If atom is not a new-style virtual
    or it does not match an installed package then it is
    yielded without any expansion.
    """
    if not isinstance(atom, Atom):
        atom = Atom(atom)

    if not atom.cp.startswith("virtual/"):
        yield atom
        return

    traversed = set()
    stack = [atom]

    while stack:
        atom = stack.pop()
        if atom.blocker or not atom.cp.startswith("virtual/"):
            yield atom
            continue

        matches = vardb.match(atom)
        if not (matches and matches[-1].startswith("virtual/")):
            yield atom
            continue

        virt_cpv = matches[-1]
        if virt_cpv in traversed:
            continue

        traversed.add(virt_cpv)
        eapi, iuse, rdepend, use = vardb.aux_get(
            virt_cpv, ["EAPI", "IUSE", "RDEPEND", "USE"]
        )
        if not portage.eapi_is_supported(eapi):
            yield atom
            continue

        eapi_attrs = _get_eapi_attrs(eapi)
        # Validate IUSE and IUSE, for early detection of vardb corruption.
        useflag_re = _get_useflag_re(eapi)
        valid_iuse = []
        for x in iuse.split():
            if x[:1] in ("+", "-"):
                x = x[1:]
            if useflag_re.match(x) is not None:
                valid_iuse.append(x)
        valid_iuse = frozenset(valid_iuse)

        if eapi_attrs.iuse_effective:
            iuse_implicit_match = vardb.settings._iuse_effective_match
        else:
            iuse_implicit_match = vardb.settings._iuse_implicit_match

        valid_use = []
        for x in use.split():
            if x in valid_iuse or iuse_implicit_match(x):
                valid_use.append(x)
        valid_use = frozenset(valid_use)

        success, atoms = portage.dep_check(
            rdepend,
            None,
            vardb.settings,
            myuse=valid_use,
            myroot=vardb.settings["EROOT"],
            trees={
                vardb.settings["EROOT"]: {
                    "porttree": vardb.vartree,
                    "vartree": vardb.vartree,
                }
            },
        )

        if success:
            stack.extend(atoms)
        else:
            yield atom



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2


class ContentsCaseSensitivityManager:
    """
    Implicitly handles case transformations that are needed for
    case-insensitive support.
    """

    def __init__(self, db):
        """
        @param db: A dblink instance
        @type db: vartree.dblink
        """
        self.getcontents = db.getcontents

        if "case-insensitive-fs" in db.settings.features:
            self.unmap_key = self._unmap_key_case_insensitive
            self.contains = self._contains_case_insensitive
            self.keys = self._keys_case_insensitive

        self._contents_insensitive = None
        self._reverse_key_map = None

    def clear_cache(self):
        """
        Clear all cached contents data.
        """
        self._contents_insensitive = None
        self._reverse_key_map = None

    def keys(self):
        """
        Iterate over all contents keys, which are transformed to
        lowercase when appropriate, for use in case-insensitive
        comparisons.
        @rtype: iterator
        @return: An iterator over all the contents keys
        """
        return iter(self.getcontents())

    def contains(self, key):
        """
        Check if the given key is contained in the contents, using
        case-insensitive comparison when appropriate.
        @param key: A filesystem path (including ROOT and EPREFIX)
        @type key: str
        @rtype: bool
        @return: True if the given key is contained in the contents,
                False otherwise
        """
        return key in self.getcontents()

    def unmap_key(self, key):
        """
        Map a key (from the keys method) back to its case-preserved
        form.
        @param key: A filesystem path (including ROOT and EPREFIX)
        @type key: str
        @rtype: str
        @return: The case-preserved form of key
        """
        return key

    def _case_insensitive_init(self):
        """
        Initialize data structures for case-insensitive support.
        """
        self._contents_insensitive = dict(
            (k.lower(), v) for k, v in self.getcontents().items()
        )
        self._reverse_key_map = dict((k.lower(), k) for k in self.getcontents())

    def _keys_case_insensitive(self):
        if self._contents_insensitive is None:
            self._case_insensitive_init()
        return iter(self._contents_insensitive)

    _keys_case_insensitive.__doc__ = keys.__doc__

    def _contains_case_insensitive(self, key):
        if self._contents_insensitive is None:
            self._case_insensitive_init()
        return key.lower() in self._contents_insensitive

    _contains_case_insensitive.__doc__ = contains.__doc__

    def _unmap_key_case_insensitive(self, key):
        if self._reverse_key_map is None:
            self._case_insensitive_init()
        return self._reverse_key_map[key]

    _unmap_key_case_insensitive.__doc__ = unmap_key.__doc__



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class CircularDependencyTestCase(TestCase):

    # TODO:
    # 	use config change by autounmask
    # 	conflict on parent's parent
    # 	difference in RDEPEND and DEPEND
    # 	is there anything else than priority buildtime and runtime?
    # 	play with use.{mask,force}
    # 	play with REQUIRED_USE

    def testCircularDependency(self):

        ebuilds = {
            "dev-libs/Z-1": {
                "DEPEND": "foo? ( !bar? ( dev-libs/Y ) )",
                "IUSE": "+foo bar",
                "EAPI": 1,
            },
            "dev-libs/Z-2": {
                "DEPEND": "foo? ( dev-libs/Y ) !bar? ( dev-libs/Y )",
                "IUSE": "+foo bar",
                "EAPI": 1,
            },
            "dev-libs/Z-3": {
                "DEPEND": "foo? ( !bar? ( dev-libs/Y ) ) foo? ( dev-libs/Y ) !bar? ( dev-libs/Y )",
                "IUSE": "+foo bar",
                "EAPI": 1,
            },
            "dev-libs/Y-1": {"DEPEND": "dev-libs/Z"},
            "dev-libs/W-1": {"DEPEND": "dev-libs/Z[foo] dev-libs/Y", "EAPI": 2},
            "dev-libs/W-2": {
                "DEPEND": "dev-libs/Z[foo=] dev-libs/Y",
                "IUSE": "+foo",
                "EAPI": 2,
            },
            "dev-libs/W-3": {"DEPEND": "dev-libs/Z[bar] dev-libs/Y", "EAPI": 2},
            "app-misc/A-1": {
                "DEPEND": "foo? ( =app-misc/B-1 )",
                "IUSE": "+foo bar",
                "REQUIRED_USE": "^^ ( foo bar )",
                "EAPI": "4",
            },
            "app-misc/A-2": {
                "DEPEND": "foo? ( =app-misc/B-2 ) bar? ( =app-misc/B-2 )",
                "IUSE": "+foo bar",
                "REQUIRED_USE": "^^ ( foo bar )",
                "EAPI": "4",
            },
            "app-misc/B-1": {"DEPEND": "=app-misc/A-1"},
            "app-misc/B-2": {"DEPEND": "=app-misc/A-2"},
        }

        test_cases = (
            # Simple tests
            ResolverPlaygroundTestCase(
                ["=dev-libs/Z-1"],
                circular_dependency_solutions={
                    "dev-libs/Y-1": frozenset(
                        [frozenset([("foo", False)]), frozenset([("bar", True)])]
                    )
                },
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/Z-2"],
                circular_dependency_solutions={
                    "dev-libs/Y-1": frozenset(
                        [frozenset([("foo", False), ("bar", True)])]
                    )
                },
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/Z-3"],
                circular_dependency_solutions={
                    "dev-libs/Y-1": frozenset(
                        [frozenset([("foo", False), ("bar", True)])]
                    )
                },
                success=False,
            ),
            # Conflict on parent
            ResolverPlaygroundTestCase(
                ["=dev-libs/W-1"], circular_dependency_solutions={}, success=False
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/W-2"],
                circular_dependency_solutions={
                    "dev-libs/Y-1": frozenset(
                        [frozenset([("foo", False), ("bar", True)])]
                    )
                },
                success=False,
            ),
            # Conflict with autounmask
            ResolverPlaygroundTestCase(
                ["=dev-libs/W-3"],
                circular_dependency_solutions={
                    "dev-libs/Y-1": frozenset([frozenset([("foo", False)])])
                },
                use_changes={"dev-libs/Z-3": {"bar": True}},
                success=False,
            ),
            # Conflict with REQUIRED_USE
            ResolverPlaygroundTestCase(
                ["=app-misc/B-1"],
                circular_dependency_solutions={
                    "app-misc/B-1": frozenset(
                        [frozenset([("foo", False), ("bar", True)])]
                    )
                },
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["=app-misc/B-2"], circular_dependency_solutions={}, success=False
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

module_metadata = {}

script_metadata = {
    "socks5-server.py": {
        "required_python": "3.3",
    },
}



# Copyright 1999-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.DependencyArg import DependencyArg
from portage._sets import SETPREFIX


class SetArg(DependencyArg):

    __slots__ = ("name", "pset")

    def __init__(self, pset=None, **kwargs):
        DependencyArg.__init__(self, **kwargs)
        self.pset = pset
        self.name = self.arg[len(SETPREFIX) :]



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class OrDowngradeInstalledTestCase(TestCase):
    def testOrDowngradeInstalled(self):
        ebuilds = {
            "net-misc/foo-1": {
                "EAPI": "6",
                "RDEPEND": "|| ( sys-libs/glibc[rpc(-)]  net-libs/libtirpc )",
            },
            "net-libs/libtirpc-1": {
                "EAPI": "6",
            },
            "sys-libs/glibc-2.26": {"EAPI": "6", "IUSE": ""},
            "sys-libs/glibc-2.24": {"EAPI": "6", "IUSE": "+rpc"},
        }

        installed = {
            "sys-libs/glibc-2.26": {"EAPI": "6", "IUSE": ""},
        }

        world = ["sys-libs/glibc"]

        test_cases = (
            # Test bug 635540, where we need to install libtirpc
            # rather than downgrade glibc.
            ResolverPlaygroundTestCase(
                ["net-misc/foo"],
                success=True,
                mergelist=[
                    "net-libs/libtirpc-1",
                    "net-misc/foo-1",
                ],
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed, world=world
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

        # In some cases it's necessary to downgrade due to
        # the installed package being masked (glibc is a
        # not an ideal example because it's usually not
        # practical to downgrade it).
        user_config = {
            "package.mask": (">=sys-libs/glibc-2.26",),
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["net-misc/foo"],
                success=True,
                mergelist=[
                    "sys-libs/glibc-2.24",
                    "net-misc/foo-1",
                ],
            ),
        )

        playground = ResolverPlayground(
            debug=False,
            ebuilds=ebuilds,
            installed=installed,
            world=world,
            user_config=user_config,
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import subprocess

try:
    import threading
except ImportError:
    # dummy_threading will not suffice
    threading = None

from portage import os
from portage.tests import TestCase
from portage.util._async.PopenProcess import PopenProcess
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util._async.PipeReaderBlockingIO import PipeReaderBlockingIO


class PopenPipeBlockingIOTestCase(TestCase):
    """
    Test PopenProcess, which can be useful for Jython support:
            * use subprocess.Popen since Jython does not support os.fork()
            * use blocking IO with threads, since Jython does not support
              fcntl non-blocking IO)
    """

    _echo_cmd = "echo -n '%s'"

    def _testPipeReader(self, test_string):
        """
        Use a poll loop to read data from a pipe and assert that
        the data written to the pipe is identical to the data
        read from the pipe.
        """

        producer = PopenProcess(
            proc=subprocess.Popen(
                ["bash", "-c", self._echo_cmd % test_string],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            ),
            pipe_reader=PipeReaderBlockingIO(),
            scheduler=global_event_loop(),
        )

        consumer = producer.pipe_reader
        consumer.input_files = {"producer": producer.proc.stdout}

        producer.start()
        producer.wait()

        self.assertEqual(producer.returncode, os.EX_OK)
        self.assertEqual(consumer.returncode, os.EX_OK)

        return consumer.getvalue().decode("ascii", "replace")

    def testPopenPipeBlockingIO(self):

        if threading is None:
            skip_reason = "threading disabled"
            self.portage_skip = "threading disabled"
            self.assertFalse(True, skip_reason)
            return

        for x in (1, 2, 5, 6, 7, 8, 2**5, 2**10, 2**12, 2**13, 2**14):
            test_string = x * "a"
            output = self._testPipeReader(test_string)
            self.assertEqual(
                test_string, output, "x = %s, len(output) = %s" % (x, len(output))
            )



# Copyright 2018-2021 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import functools

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.util.futures:asyncio",
)

# A marker for iscoroutinefunction.
_is_coroutine = object()


def _iscoroutinefunction(func):
    """
    Return True if func is a decorated coroutine function
    created with the coroutine decorator for this module.
    """
    return getattr(func, "_is_coroutine", None) is _is_coroutine


def coroutine(generator_func):
    """
    A decorator for a generator function that behaves as coroutine function.
    The generator should yield a Future instance in order to wait for it,
    and the result becomes the result of the current yield-expression,
    via the PEP 342 generator send() method.

    The decorated function returns a Future which is done when the generator
    is exhausted. The generator can return a value via the coroutine_return
    function.

    @param generator_func: A generator function that yields Futures, and
            will receive the result of each Future as the result of the
            corresponding yield-expression.
    @type generator_func: function
    @rtype: function
    @return: A function which calls the given generator function and
            returns a Future that is done when the generator is exhausted.
    """
    # Note that functools.partial does not work for decoration of
    # methods, since it doesn't implement the descriptor protocol.
    # This problem is solve by defining a wrapper function.
    @functools.wraps(generator_func)
    def wrapped(*args, **kwargs):
        return _generator_future(generator_func, *args, **kwargs)

    wrapped._is_coroutine = _is_coroutine
    return wrapped


def coroutine_return(result=None):
    """
    Terminate the current coroutine and set the result of the associated
    Future.

    @param result: of the current coroutine's Future
    @type object
    """
    raise _CoroutineReturnValue(result)


def _generator_future(generator_func, *args, **kwargs):
    """
    Call generator_func with the given arguments, and return a Future
    that is done when the resulting generation is exhausted. If a
    keyword argument named 'loop' is given, then it is used instead of
    the default event loop.
    """
    loop = kwargs.get("loop")
    loop = asyncio._wrap_loop(loop)
    result = loop.create_future()
    _GeneratorTask(generator_func(*args, **kwargs), result, loop=loop)
    return result


class _CoroutineReturnValue(Exception):
    def __init__(self, result):
        self.result = result


class _GeneratorTask:
    """
    Asynchronously executes the generator to completion, waiting for
    the result of each Future that it yields, and sending the result
    to the generator.
    """

    def __init__(self, generator, result, loop):
        self._generator = generator
        self._result = result
        self._current_task = None
        self._loop = loop
        result.add_done_callback(self._cancel_callback)
        loop.call_soon(self._next)

    def _cancel_callback(self, result):
        if result.cancelled() and self._current_task is not None:
            # The done callback for self._current_task invokes
            # _next in either case here.
            self._current_task.done() or self._current_task.cancel()

    def _next(self, previous=None):
        self._current_task = None
        if self._result.cancelled():
            if previous is not None:
                # Consume exceptions, in order to avoid triggering
                # the event loop's exception handler.
                previous.cancelled() or previous.exception()

            # This will throw asyncio.CancelledError in the coroutine if
            # there's an opportunity (yield) before the generator raises
            # StopIteration.
            previous = self._result
        try:
            if previous is None:
                future = next(self._generator)
            elif previous.cancelled():
                future = self._generator.throw(asyncio.CancelledError())
            elif previous.exception() is None:
                future = self._generator.send(previous.result())
            else:
                future = self._generator.throw(previous.exception())

        except asyncio.CancelledError:
            self._result.cancel()
        except _CoroutineReturnValue as e:
            if not self._result.cancelled():
                self._result.set_result(e.result)
        except StopIteration:
            if not self._result.cancelled():
                self._result.set_result(None)
        except Exception as e:
            if not self._result.cancelled():
                self._result.set_exception(e)
        else:
            self._current_task = asyncio.ensure_future(future, loop=self._loop)
            self._current_task.add_done_callback(self._next)



# Copyright 2009-2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.exception import InvalidDependString
from portage.dep import Atom, use_reduce


class UseReduceTestCase:
    def __init__(
        self,
        deparray,
        uselist=[],
        masklist=[],
        matchall=0,
        excludeall=[],
        is_src_uri=False,
        eapi="0",
        opconvert=False,
        flat=False,
        expected_result=None,
        is_valid_flag=None,
        token_class=None,
        subset=None,
    ):
        self.deparray = deparray
        self.uselist = uselist
        self.masklist = masklist
        self.matchall = matchall
        self.excludeall = excludeall
        self.is_src_uri = is_src_uri
        self.eapi = eapi
        self.opconvert = opconvert
        self.flat = flat
        self.is_valid_flag = is_valid_flag
        self.token_class = token_class
        self.subset = subset
        self.expected_result = expected_result

    def run(self):
        try:
            return use_reduce(
                self.deparray,
                self.uselist,
                self.masklist,
                self.matchall,
                self.excludeall,
                self.is_src_uri,
                self.eapi,
                self.opconvert,
                self.flat,
                self.is_valid_flag,
                self.token_class,
                subset=self.subset,
            )
        except InvalidDependString as e:
            raise InvalidDependString("%s: %s" % (e, self.deparray))


class UseReduce(TestCase):
    def always_true(self, ununsed_parameter):
        return True

    def always_false(self, ununsed_parameter):
        return False

    def testUseReduce(self):

        EAPI_WITH_SRC_URI_ARROWS = "2"
        EAPI_WITHOUT_SRC_URI_ARROWS = "0"

        test_cases = (
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                uselist=["a", "b", "c", "d"],
                expected_result=["A", "B"],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                uselist=["a", "b", "c", "d"],
                subset=["b"],
                expected_result=["B"],
            ),
            UseReduceTestCase(
                "bar? ( || ( foo bar? ( baz ) ) )",
                uselist=["bar"],
                subset=["bar"],
                expected_result=["||", ["foo", "baz"]],
            ),
            UseReduceTestCase(
                "bar? ( foo bar? ( baz ) foo )",
                uselist=["bar"],
                subset=["bar"],
                expected_result=["foo", "baz", "foo"],
            ),
            UseReduceTestCase(
                "|| ( ( a b ) ( c d ) )", uselist=[], subset=["bar"], expected_result=[]
            ),
            UseReduceTestCase(
                "|| ( ( a b ) ( bar? ( c d ) e f ) )",
                uselist=["bar"],
                subset=["bar"],
                expected_result=["c", "d"],
            ),
            UseReduceTestCase(
                "( a b ) ( c d bar? ( e f baz? ( g h ) ) )",
                uselist=["bar"],
                subset=["bar"],
                expected_result=["e", "f"],
            ),
            UseReduceTestCase(
                "( a b ) ( c d bar? ( e f baz? ( g h ) ) )",
                uselist=["bar", "baz"],
                subset=["bar"],
                expected_result=["e", "f", "g", "h"],
            ),
            UseReduceTestCase(
                "( bar? ( a b ) ( bar? ( c d ) ) ) ( e f )",
                uselist=["bar"],
                subset=["bar"],
                expected_result=["a", "b", "c", "d"],
            ),
            UseReduceTestCase(
                "|| ( foo bar? ( baz ) )",
                uselist=["bar"],
                subset=["bar"],
                expected_result=["baz"],
            ),
            UseReduceTestCase(
                "|| ( || ( bar? ( a || ( b c || ( d e ) ) ) ) )",
                uselist=["bar"],
                subset=["bar"],
                expected_result=["a", "||", ["b", "c", "d", "e"]],
            ),
            UseReduceTestCase(
                "|| ( || ( bar? ( a || ( ( b c ) ( d e ) ) ) ) )",
                uselist=["bar"],
                subset=["bar"],
                expected_result=["a", "||", [["b", "c"], ["d", "e"]]],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                uselist=["a", "b", "c"],
                expected_result=["A", "B", "D"],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                uselist=["b", "c"],
                expected_result=["B", "D"],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                matchall=True,
                expected_result=["A", "B", "C", "D"],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                masklist=["a", "c"],
                expected_result=["C", "D"],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                matchall=True,
                masklist=["a", "c"],
                expected_result=["B", "C", "D"],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                uselist=["a", "b"],
                masklist=["a", "c"],
                expected_result=["B", "C", "D"],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                excludeall=["a", "c"],
                expected_result=["D"],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                uselist=["b"],
                excludeall=["a", "c"],
                expected_result=["B", "D"],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                matchall=True,
                excludeall=["a", "c"],
                expected_result=["A", "B", "D"],
            ),
            UseReduceTestCase(
                "a? ( A ) b? ( B ) !c? ( C ) !d? ( D )",
                matchall=True,
                excludeall=["a", "c"],
                masklist=["b"],
                expected_result=["A", "D"],
            ),
            UseReduceTestCase(
                "a? ( b? ( AB ) )", uselist=["a", "b"], expected_result=["AB"]
            ),
            UseReduceTestCase(
                "a? ( b? ( AB ) C )", uselist=["a"], expected_result=["C"]
            ),
            UseReduceTestCase(
                "a? ( b? ( || ( AB CD ) ) )",
                uselist=["a", "b"],
                expected_result=["||", ["AB", "CD"]],
            ),
            UseReduceTestCase(
                "|| ( || ( a? ( A ) b? ( B ) ) )",
                uselist=["a", "b"],
                expected_result=["||", ["A", "B"]],
            ),
            UseReduceTestCase(
                "|| ( || ( a? ( A ) b? ( B ) ) )", uselist=["a"], expected_result=["A"]
            ),
            UseReduceTestCase(
                "|| ( || ( a? ( A ) b? ( B ) ) )", uselist=[], expected_result=[]
            ),
            UseReduceTestCase(
                "|| ( || ( a? ( || ( A c? ( C ) ) ) b? ( B ) ) )",
                uselist=[],
                expected_result=[],
            ),
            UseReduceTestCase(
                "|| ( || ( a? ( || ( A c? ( C ) ) ) b? ( B ) ) )",
                uselist=["a"],
                expected_result=["A"],
            ),
            UseReduceTestCase(
                "|| ( || ( a? ( || ( A c? ( C ) ) ) b? ( B ) ) )",
                uselist=["b"],
                expected_result=["B"],
            ),
            UseReduceTestCase(
                "|| ( || ( a? ( || ( A c? ( C ) ) ) b? ( B ) ) )",
                uselist=["c"],
                expected_result=[],
            ),
            UseReduceTestCase(
                "|| ( || ( a? ( || ( A c? ( C ) ) ) b? ( B ) ) )",
                uselist=["a", "c"],
                expected_result=["||", ["A", "C"]],
            ),
            # paren_reduce tests
            UseReduceTestCase("A", expected_result=["A"]),
            UseReduceTestCase("( A )", expected_result=["A"]),
            UseReduceTestCase("|| ( A B )", expected_result=["||", ["A", "B"]]),
            UseReduceTestCase(
                "|| ( ( A B ) C )", expected_result=["||", [["A", "B"], "C"]]
            ),
            UseReduceTestCase(
                "|| ( ( A B ) ( C ) )", expected_result=["||", [["A", "B"], "C"]]
            ),
            UseReduceTestCase(
                "|| ( A || ( B C ) )", expected_result=["||", ["A", "B", "C"]]
            ),
            UseReduceTestCase(
                "|| ( A || ( B C D ) )", expected_result=["||", ["A", "B", "C", "D"]]
            ),
            UseReduceTestCase(
                "|| ( A || ( B || ( C D ) E ) )",
                expected_result=["||", ["A", "B", "C", "D", "E"]],
            ),
            UseReduceTestCase("( || ( ( ( A ) B ) ) )", expected_result=["A", "B"]),
            UseReduceTestCase(
                "( || ( || ( ( A ) B ) ) )", expected_result=["||", ["A", "B"]]
            ),
            UseReduceTestCase(
                "( || ( || ( ( A ) B ) ) )", expected_result=["||", ["A", "B"]]
            ),
            UseReduceTestCase("|| ( A )", expected_result=["A"]),
            UseReduceTestCase(
                "( || ( || ( || ( A ) foo? ( B ) ) ) )", expected_result=["A"]
            ),
            UseReduceTestCase(
                "( || ( || ( || ( A ) foo? ( B ) ) ) )",
                uselist=["foo"],
                expected_result=["||", ["A", "B"]],
            ),
            UseReduceTestCase(
                "( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )", expected_result=[]
            ),
            UseReduceTestCase(
                "( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
                uselist=["foo", "bar"],
                expected_result=["||", ["A", "B"]],
            ),
            UseReduceTestCase(
                "A || ( bar? ( C ) ) foo? ( bar? ( C ) ) B", expected_result=["A", "B"]
            ),
            UseReduceTestCase("|| ( A ) || ( B )", expected_result=["A", "B"]),
            UseReduceTestCase("foo? ( A ) foo? ( B )", expected_result=[]),
            UseReduceTestCase(
                "foo? ( A ) foo? ( B )", uselist=["foo"], expected_result=["A", "B"]
            ),
            UseReduceTestCase("|| ( A B ) C", expected_result=["||", ["A", "B"], "C"]),
            UseReduceTestCase("A || ( B C )", expected_result=["A", "||", ["B", "C"]]),
            # SRC_URI stuff
            UseReduceTestCase(
                "http://foo/bar -> blah.tbz2",
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
                expected_result=["http://foo/bar", "->", "blah.tbz2"],
            ),
            UseReduceTestCase(
                "foo? ( http://foo/bar -> blah.tbz2 )",
                uselist=[],
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
                expected_result=[],
            ),
            UseReduceTestCase(
                "foo? ( http://foo/bar -> blah.tbz2 )",
                uselist=["foo"],
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
                expected_result=["http://foo/bar", "->", "blah.tbz2"],
            ),
            UseReduceTestCase(
                "http://foo/bar -> bar.tbz2 foo? ( ftp://foo/a )",
                uselist=[],
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
                expected_result=["http://foo/bar", "->", "bar.tbz2"],
            ),
            UseReduceTestCase(
                "http://foo/bar -> bar.tbz2 foo? ( ftp://foo/a )",
                uselist=["foo"],
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
                expected_result=["http://foo/bar", "->", "bar.tbz2", "ftp://foo/a"],
            ),
            UseReduceTestCase(
                "http://foo.com/foo http://foo/bar -> blah.tbz2",
                uselist=["foo"],
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
                expected_result=[
                    "http://foo.com/foo",
                    "http://foo/bar",
                    "->",
                    "blah.tbz2",
                ],
            ),
            # opconvert tests
            UseReduceTestCase("A", opconvert=True, expected_result=["A"]),
            UseReduceTestCase("( A )", opconvert=True, expected_result=["A"]),
            UseReduceTestCase(
                "|| ( A B )", opconvert=True, expected_result=[["||", "A", "B"]]
            ),
            UseReduceTestCase(
                "|| ( ( A B ) C )",
                opconvert=True,
                expected_result=[["||", ["A", "B"], "C"]],
            ),
            UseReduceTestCase(
                "|| ( A || ( B C ) )",
                opconvert=True,
                expected_result=[["||", "A", "B", "C"]],
            ),
            UseReduceTestCase(
                "|| ( A || ( B C D ) )",
                opconvert=True,
                expected_result=[["||", "A", "B", "C", "D"]],
            ),
            UseReduceTestCase(
                "|| ( A || ( B || ( C D ) E ) )",
                expected_result=["||", ["A", "B", "C", "D", "E"]],
            ),
            UseReduceTestCase(
                "( || ( ( ( A ) B ) ) )", opconvert=True, expected_result=["A", "B"]
            ),
            UseReduceTestCase(
                "( || ( || ( ( A ) B ) ) )",
                opconvert=True,
                expected_result=[["||", "A", "B"]],
            ),
            UseReduceTestCase(
                "|| ( A B ) C", opconvert=True, expected_result=[["||", "A", "B"], "C"]
            ),
            UseReduceTestCase(
                "A || ( B C )", opconvert=True, expected_result=["A", ["||", "B", "C"]]
            ),
            UseReduceTestCase(
                "A foo? ( || ( B || ( bar? ( || ( C D E ) ) !bar? ( F ) ) ) ) G",
                uselist=["foo", "bar"],
                opconvert=True,
                expected_result=["A", ["||", "B", "C", "D", "E"], "G"],
            ),
            UseReduceTestCase(
                "A foo? ( || ( B || ( bar? ( || ( C D E ) ) !bar? ( F ) ) ) ) G",
                uselist=["foo", "bar"],
                opconvert=False,
                expected_result=["A", "||", ["B", "C", "D", "E"], "G"],
            ),
            UseReduceTestCase("|| ( A )", opconvert=True, expected_result=["A"]),
            UseReduceTestCase(
                "( || ( || ( || ( A ) foo? ( B ) ) ) )", expected_result=["A"]
            ),
            UseReduceTestCase(
                "( || ( || ( || ( A ) foo? ( B ) ) ) )",
                uselist=["foo"],
                opconvert=True,
                expected_result=[["||", "A", "B"]],
            ),
            UseReduceTestCase(
                "( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
                opconvert=True,
                expected_result=[],
            ),
            UseReduceTestCase(
                "( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
                uselist=["foo", "bar"],
                opconvert=True,
                expected_result=[["||", "A", "B"]],
            ),
            UseReduceTestCase(
                "A || ( bar? ( C ) ) foo? ( bar? ( C ) ) B",
                opconvert=True,
                expected_result=["A", "B"],
            ),
            UseReduceTestCase(
                "|| ( A ) || ( B )", opconvert=True, expected_result=["A", "B"]
            ),
            UseReduceTestCase(
                "foo? ( A ) foo? ( B )", opconvert=True, expected_result=[]
            ),
            UseReduceTestCase(
                "foo? ( A ) foo? ( B )",
                uselist=["foo"],
                opconvert=True,
                expected_result=["A", "B"],
            ),
            UseReduceTestCase(
                "|| ( foo? ( || ( A B ) ) )",
                uselist=["foo"],
                opconvert=True,
                expected_result=[["||", "A", "B"]],
            ),
            UseReduceTestCase(
                "|| ( ( A B ) foo? ( || ( C D ) ) )",
                uselist=["foo"],
                opconvert=True,
                expected_result=[["||", ["A", "B"], "C", "D"]],
            ),
            UseReduceTestCase(
                "|| ( ( A B ) foo? ( || ( C D ) ) )",
                uselist=["foo"],
                opconvert=False,
                expected_result=["||", [["A", "B"], "C", "D"]],
            ),
            UseReduceTestCase(
                "|| ( ( A B ) || ( C D ) )",
                expected_result=["||", [["A", "B"], "C", "D"]],
            ),
            UseReduceTestCase(
                "|| ( ( A B ) || ( C D || ( E ( F G ) || ( H ) ) ) )",
                expected_result=["||", [["A", "B"], "C", "D", "E", ["F", "G"], "H"]],
            ),
            UseReduceTestCase(
                "|| ( ( A B ) || ( C D || ( E ( F G ) || ( H ) ) ) )",
                opconvert=True,
                expected_result=[["||", ["A", "B"], "C", "D", "E", ["F", "G"], "H"]],
            ),
            UseReduceTestCase(
                "|| ( foo? ( A B ) )", uselist=["foo"], expected_result=["A", "B"]
            ),
            UseReduceTestCase(
                "|| ( || ( foo? ( A B ) ) )",
                uselist=["foo"],
                expected_result=["A", "B"],
            ),
            UseReduceTestCase(
                "|| ( || ( || ( a? ( b? ( c? ( || ( || ( || ( d? ( e? ( f? ( A B ) ) ) ) ) ) ) ) ) ) ) )",
                uselist=["a", "b", "c", "d", "e", "f"],
                expected_result=["A", "B"],
            ),
            UseReduceTestCase(
                "|| ( || ( ( || ( a? ( ( b? ( c? ( || ( || ( || ( ( d? ( e? ( f? ( A B ) ) ) ) ) ) ) ) ) ) ) ) ) ) )",
                uselist=["a", "b", "c", "d", "e", "f"],
                expected_result=["A", "B"],
            ),
            UseReduceTestCase("|| ( ( A ( || ( B ) ) ) )", expected_result=["A", "B"]),
            UseReduceTestCase(
                "|| ( ( A B ) || ( foo? ( bar? ( ( C D || ( baz? ( E ) ( F G ) || ( H ) ) ) ) ) ) )",
                uselist=["foo", "bar", "baz"],
                expected_result=[
                    "||",
                    [["A", "B"], ["C", "D", "||", ["E", ["F", "G"], "H"]]],
                ],
            ),
            UseReduceTestCase(
                "|| ( ( A B ) || ( foo? ( bar? ( ( C D || ( baz? ( E ) ( F G ) || ( H ) ) ) ) ) ) )",
                uselist=["foo", "bar", "baz"],
                opconvert=True,
                expected_result=[
                    ["||", ["A", "B"], ["C", "D", ["||", "E", ["F", "G"], "H"]]]
                ],
            ),
            UseReduceTestCase(
                "|| ( foo? ( A B ) )",
                uselist=["foo"],
                opconvert=True,
                expected_result=["A", "B"],
            ),
            UseReduceTestCase(
                "|| ( || ( foo? ( A B ) ) )",
                uselist=["foo"],
                opconvert=True,
                expected_result=["A", "B"],
            ),
            UseReduceTestCase(
                "|| ( || ( || ( a? ( b? ( c? ( || ( || ( || ( d? ( e? ( f? ( A B ) ) ) ) ) ) ) ) ) ) ) )",
                uselist=["a", "b", "c", "d", "e", "f"],
                opconvert=True,
                expected_result=["A", "B"],
            ),
            # flat test
            UseReduceTestCase("A", flat=True, expected_result=["A"]),
            UseReduceTestCase("( A )", flat=True, expected_result=["A"]),
            UseReduceTestCase(
                "|| ( A B )", flat=True, expected_result=["||", "A", "B"]
            ),
            UseReduceTestCase(
                "|| ( A || ( B C ) )",
                flat=True,
                expected_result=["||", "A", "||", "B", "C"],
            ),
            UseReduceTestCase(
                "|| ( A || ( B C D ) )",
                flat=True,
                expected_result=["||", "A", "||", "B", "C", "D"],
            ),
            UseReduceTestCase(
                "|| ( A || ( B || ( C D ) E ) )",
                flat=True,
                expected_result=["||", "A", "||", "B", "||", "C", "D", "E"],
            ),
            UseReduceTestCase(
                "( || ( ( ( A ) B ) ) )", flat=True, expected_result=["||", "A", "B"]
            ),
            UseReduceTestCase(
                "( || ( || ( ( A ) B ) ) )",
                flat=True,
                expected_result=["||", "||", "A", "B"],
            ),
            UseReduceTestCase(
                "( || ( || ( ( A ) B ) ) )",
                flat=True,
                expected_result=["||", "||", "A", "B"],
            ),
            UseReduceTestCase("|| ( A )", flat=True, expected_result=["||", "A"]),
            UseReduceTestCase(
                "( || ( || ( || ( A ) foo? ( B ) ) ) )", expected_result=["A"]
            ),
            UseReduceTestCase(
                "( || ( || ( || ( A ) foo? ( B ) ) ) )",
                uselist=["foo"],
                flat=True,
                expected_result=["||", "||", "||", "A", "B"],
            ),
            UseReduceTestCase(
                "( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
                flat=True,
                expected_result=["||", "||", "||"],
            ),
            UseReduceTestCase(
                "( || ( || ( bar? ( A ) || ( foo? ( B ) ) ) ) )",
                uselist=["foo", "bar"],
                flat=True,
                expected_result=["||", "||", "A", "||", "B"],
            ),
            UseReduceTestCase(
                "A || ( bar? ( C ) ) foo? ( bar? ( C ) ) B",
                flat=True,
                expected_result=["A", "||", "B"],
            ),
            UseReduceTestCase(
                "|| ( A ) || ( B )", flat=True, expected_result=["||", "A", "||", "B"]
            ),
            UseReduceTestCase("foo? ( A ) foo? ( B )", flat=True, expected_result=[]),
            UseReduceTestCase(
                "foo? ( A ) foo? ( B )",
                uselist=["foo"],
                flat=True,
                expected_result=["A", "B"],
            ),
            # use flag validation
            UseReduceTestCase(
                "foo? ( A )",
                uselist=["foo"],
                is_valid_flag=self.always_true,
                expected_result=["A"],
            ),
            UseReduceTestCase(
                "foo? ( A )", is_valid_flag=self.always_true, expected_result=[]
            ),
            # token_class
            UseReduceTestCase(
                "foo? ( dev-libs/A )",
                uselist=["foo"],
                token_class=Atom,
                expected_result=["dev-libs/A"],
            ),
            UseReduceTestCase(
                "foo? ( dev-libs/A )", token_class=Atom, expected_result=[]
            ),
        )

        test_cases_xfail = (
            UseReduceTestCase("? ( A )"),
            UseReduceTestCase("!? ( A )"),
            UseReduceTestCase("( A"),
            UseReduceTestCase("A )"),
            UseReduceTestCase("||( A B )"),
            UseReduceTestCase("|| (A B )"),
            UseReduceTestCase("|| ( A B)"),
            UseReduceTestCase("|| ( A B"),
            UseReduceTestCase("|| A B )"),
            UseReduceTestCase("|| A B"),
            UseReduceTestCase("|| ( A B ) )"),
            UseReduceTestCase("|| || B C"),
            UseReduceTestCase("|| ( A B || )"),
            UseReduceTestCase("a? A"),
            UseReduceTestCase("( || ( || || ( A ) foo? ( B ) ) )"),
            UseReduceTestCase("( || ( || bar? ( A ) foo? ( B ) ) )"),
            UseReduceTestCase("foo?"),
            UseReduceTestCase("foo? || ( A )"),
            UseReduceTestCase("|| ( )"),
            UseReduceTestCase("foo? ( )"),
            # SRC_URI stuff
            UseReduceTestCase(
                "http://foo/bar -> blah.tbz2",
                is_src_uri=True,
                eapi=EAPI_WITHOUT_SRC_URI_ARROWS,
            ),
            UseReduceTestCase(
                "|| ( http://foo/bar -> blah.tbz2 )",
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
            ),
            UseReduceTestCase(
                "http://foo/bar -> foo? ( ftp://foo/a )",
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
            ),
            UseReduceTestCase(
                "http://foo/bar blah.tbz2 ->",
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
            ),
            UseReduceTestCase(
                "-> http://foo/bar blah.tbz2 )",
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
            ),
            UseReduceTestCase(
                "http://foo/bar ->", is_src_uri=True, eapi=EAPI_WITH_SRC_URI_ARROWS
            ),
            UseReduceTestCase(
                "http://foo/bar -> foo? ( http://foo.com/foo )",
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
            ),
            UseReduceTestCase(
                "foo? ( http://foo/bar -> ) blah.tbz2",
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
            ),
            UseReduceTestCase(
                "http://foo/bar -> foo/blah.tbz2",
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
            ),
            UseReduceTestCase(
                "http://foo/bar -> -> bar.tbz2 foo? ( ftp://foo/a )",
                is_src_uri=True,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
            ),
            UseReduceTestCase(
                "http://foo/bar -> bar.tbz2 foo? ( ftp://foo/a )",
                is_src_uri=False,
                eapi=EAPI_WITH_SRC_URI_ARROWS,
            ),
            UseReduceTestCase("A", opconvert=True, flat=True),
            # use flag validation
            UseReduceTestCase("1.0? ( A )"),
            UseReduceTestCase("!1.0? ( A )"),
            UseReduceTestCase("!? ( A )"),
            UseReduceTestCase("!?? ( A )"),
            UseReduceTestCase(
                "foo? ( A )",
                is_valid_flag=self.always_false,
            ),
            UseReduceTestCase(
                "foo? ( A )",
                uselist=["foo"],
                is_valid_flag=self.always_false,
            ),
            # token_class
            UseReduceTestCase("foo? ( A )", uselist=["foo"], token_class=Atom),
            UseReduceTestCase("A(B", token_class=Atom),
        )

        for test_case in test_cases:
            # If it fails then show the input, since lots of our
            # test cases have the same output but different input,
            # making it difficult deduce which test has failed.
            self.assertEqual(
                test_case.run(),
                test_case.expected_result,
                "input: '%s' result: %s != %s"
                % (test_case.deparray, test_case.run(), test_case.expected_result),
            )

        for test_case in test_cases_xfail:
            self.assertRaisesMsg(
                test_case.deparray, (InvalidDependString, ValueError), test_case.run
            )



# Copyright 2014-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.repository.config import allow_profile_repo_deps
from portage.util import grabfile_package, stack_lists
from portage._sets.base import PackageSet


class ProfilePackageSet(PackageSet):
    _operations = ["merge"]

    def __init__(self, profiles, debug=False):
        super(ProfilePackageSet, self).__init__(
            allow_repo=any(allow_profile_repo_deps(y) for y in profiles)
        )
        self._profiles = profiles
        if profiles:
            desc_profile = profiles[-1]
            if desc_profile.user_config and len(profiles) > 1:
                desc_profile = profiles[-2]
            description = desc_profile.location
        else:
            description = None
        self.description = "Profile packages for profile %s" % description

    def load(self):
        self._setAtoms(
            x
            for x in stack_lists(
                [
                    grabfile_package(
                        os.path.join(y.location, "packages"),
                        verify_eapi=True,
                        eapi=y.eapi,
                        eapi_default=None,
                        allow_build_id=y.allow_build_id,
                        allow_repo=allow_profile_repo_deps(y),
                    )
                    for y in self._profiles
                    if "profile-set" in y.profile_formats
                ],
                incremental=1,
            )
            if x[:1] != "*"
        )

    def singleBuilder(self, options, settings, trees):
        return ProfilePackageSet(settings._locations_manager.profiles_complex)

    singleBuilder = classmethod(singleBuilder)



# Copyright 2012-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from __future__ import print_function
import sys
import re
import textwrap

import portage
from portage import os
from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.dep import Atom
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.update import update_dbentry
from portage.util import ensure_dirs
from portage.versions import _pkg_str
from portage._global_updates import _do_global_updates
from portage.output import colorize


class UpdateDbentryTestCase(TestCase):
    def testUpdateDbentryTestCase(self):
        cases = (
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                "1",
                "  dev-libs/A:0  ",
                "  dev-libs/B:0  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                "1",
                "  >=dev-libs/A-1:0  ",
                "  >=dev-libs/B-1:0  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                "2",
                "  dev-libs/A[foo]  ",
                "  dev-libs/B[foo]  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                "5",
                "  dev-libs/A:0/1=[foo]  ",
                "  dev-libs/B:0/1=[foo]  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                "5",
                "  dev-libs/A:0/1[foo]  ",
                "  dev-libs/B:0/1[foo]  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                "5",
                "  dev-libs/A:0/0[foo]  ",
                "  dev-libs/B:0/0[foo]  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                "5",
                "  dev-libs/A:0=[foo]  ",
                "  dev-libs/B:0=[foo]  ",
            ),
            (
                ("slotmove", Atom("dev-libs/A"), "0", "1"),
                "1",
                "  dev-libs/A:0  ",
                "  dev-libs/A:1  ",
            ),
            (
                ("slotmove", Atom("dev-libs/A"), "0", "1"),
                "1",
                "  >=dev-libs/A-1:0  ",
                "  >=dev-libs/A-1:1  ",
            ),
            (
                ("slotmove", Atom("dev-libs/A"), "0", "1"),
                "5",
                "  dev-libs/A:0/1=[foo]  ",
                "  dev-libs/A:1/1=[foo]  ",
            ),
            (
                ("slotmove", Atom("dev-libs/A"), "0", "1"),
                "5",
                "  dev-libs/A:0/1[foo]  ",
                "  dev-libs/A:1/1[foo]  ",
            ),
            (
                ("slotmove", Atom("dev-libs/A"), "0", "1"),
                "5",
                "  dev-libs/A:0/0[foo]  ",
                "  dev-libs/A:1/1[foo]  ",
            ),
            (
                ("slotmove", Atom("dev-libs/A"), "0", "1"),
                "5",
                "  dev-libs/A:0=[foo]  ",
                "  dev-libs/A:1=[foo]  ",
            ),
        )
        for update_cmd, eapi, input_str, output_str in cases:
            result = update_dbentry(update_cmd, input_str, eapi=eapi)
            self.assertEqual(result, output_str)

    def testUpdateDbentryBlockerTestCase(self):
        """
        Avoid creating self-blockers for bug #367215.
        """
        cases = (
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                _pkg_str("dev-libs/B-1", eapi="1", slot="0"),
                "  !dev-libs/A  ",
                "  !dev-libs/A  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                _pkg_str("dev-libs/C-1", eapi="1", slot="0"),
                "  !dev-libs/A  ",
                "  !dev-libs/B  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                _pkg_str("dev-libs/B-1", eapi="1", slot="0"),
                "  !dev-libs/A:0  ",
                "  !dev-libs/A:0  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                _pkg_str("dev-libs/C-1", eapi="1", slot="0"),
                "  !dev-libs/A:0  ",
                "  !dev-libs/B:0  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                _pkg_str("dev-libs/C-1", eapi="1", slot="0"),
                "  !>=dev-libs/A-1:0  ",
                "  !>=dev-libs/B-1:0  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                _pkg_str("dev-libs/B-1", eapi="1", slot="0"),
                "  !>=dev-libs/A-1:0  ",
                "  !>=dev-libs/A-1:0  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                _pkg_str("dev-libs/C-1", eapi="1", slot="0"),
                "  !>=dev-libs/A-1  ",
                "  !>=dev-libs/B-1  ",
            ),
            (
                ("move", Atom("dev-libs/A"), Atom("dev-libs/B")),
                _pkg_str("dev-libs/B-1", eapi="1", slot="0"),
                "  !>=dev-libs/A-1  ",
                "  !>=dev-libs/A-1  ",
            ),
        )
        for update_cmd, parent, input_str, output_str in cases:
            result = update_dbentry(update_cmd, input_str, parent=parent)
            self.assertEqual(result, output_str)

    def testUpdateDbentryDbapiTestCase(self):

        ebuilds = {
            "dev-libs/A-2::dont_apply_updates": {
                "RDEPEND": "dev-libs/M dev-libs/N dev-libs/P",
                "EAPI": "4",
                "SLOT": "2",
            },
            "dev-libs/B-2::dont_apply_updates": {
                "RDEPEND": "dev-libs/M dev-libs/N dev-libs/P",
                "EAPI": "4",
                "SLOT": "2",
            },
        }

        installed = {
            "dev-libs/A-1::test_repo": {
                "RDEPEND": "dev-libs/M dev-libs/N dev-libs/P",
                "EAPI": "4",
            },
            "dev-libs/A-2::dont_apply_updates": {
                "RDEPEND": "dev-libs/M dev-libs/N dev-libs/P",
                "EAPI": "4",
                "SLOT": "2",
            },
            "dev-libs/B-1::test_repo": {
                "RDEPEND": "dev-libs/M dev-libs/N dev-libs/P",
                "EAPI": "4",
            },
            "dev-libs/M-1::test_repo": {
                "EAPI": "4",
            },
            "dev-libs/N-1::test_repo": {
                "EAPI": "4",
            },
            "dev-libs/N-2::test_repo": {
                "EAPI": "4",
            },
        }

        binpkgs = {
            "dev-libs/A-1::test_repo": {
                "RDEPEND": "dev-libs/M dev-libs/N dev-libs/P",
                "EAPI": "4",
            },
            "dev-libs/A-2::dont_apply_updates": {
                "RDEPEND": "dev-libs/M dev-libs/N dev-libs/P",
                "EAPI": "4",
                "SLOT": "2",
            },
            "dev-libs/B-1::test_repo": {
                "RDEPEND": "dev-libs/M dev-libs/N dev-libs/P",
                "EAPI": "4",
            },
        }

        world = ["dev-libs/M", "dev-libs/N"]

        updates = textwrap.dedent(
            """
			move dev-libs/M dev-libs/M-moved
		"""
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    ebuilds=ebuilds,
                    installed=installed,
                    world=world,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                settings = playground.settings
                trees = playground.trees
                eroot = settings["EROOT"]
                test_repo_location = settings.repositories["test_repo"].location
                portdb = trees[eroot]["porttree"].dbapi
                vardb = trees[eroot]["vartree"].dbapi
                bindb = trees[eroot]["bintree"].dbapi
                setconfig = trees[eroot]["root_config"].setconfig
                selected_set = setconfig.getSets()["selected"]

                updates_dir = os.path.join(test_repo_location, "profiles", "updates")

                try:
                    ensure_dirs(updates_dir)
                    with open(os.path.join(updates_dir, "1Q-2010"), "w") as f:
                        f.write(updates)

                    # Create an empty updates directory, so that this
                    # repo doesn't inherit updates from the main repo.
                    ensure_dirs(
                        os.path.join(
                            portdb.getRepositoryPath("dont_apply_updates"),
                            "profiles",
                            "updates",
                        )
                    )

                    global_noiselimit = portage.util.noiselimit
                    portage.util.noiselimit = -2
                    try:
                        _do_global_updates(trees, {})
                    finally:
                        portage.util.noiselimit = global_noiselimit

                    # Workaround for cache validation not working
                    # correctly when filesystem has timestamp precision
                    # of 1 second.
                    vardb._clear_cache()

                    # M -> M-moved
                    old_pattern = re.compile(r"\bdev-libs/M(\s|$)")
                    rdepend = vardb.aux_get("dev-libs/A-1", ["RDEPEND"])[0]
                    self.assertTrue(old_pattern.search(rdepend) is None)
                    self.assertTrue("dev-libs/M-moved" in rdepend)
                    rdepend = bindb.aux_get("dev-libs/A-1", ["RDEPEND"])[0]
                    self.assertTrue(old_pattern.search(rdepend) is None)
                    self.assertTrue("dev-libs/M-moved" in rdepend)
                    rdepend = vardb.aux_get("dev-libs/B-1", ["RDEPEND"])[0]
                    self.assertTrue(old_pattern.search(rdepend) is None)
                    self.assertTrue("dev-libs/M-moved" in rdepend)
                    rdepend = vardb.aux_get("dev-libs/B-1", ["RDEPEND"])[0]
                    self.assertTrue(old_pattern.search(rdepend) is None)
                    self.assertTrue("dev-libs/M-moved" in rdepend)

                    # dont_apply_updates
                    rdepend = vardb.aux_get("dev-libs/A-2", ["RDEPEND"])[0]
                    self.assertTrue("dev-libs/M" in rdepend)
                    self.assertTrue("dev-libs/M-moved" not in rdepend)
                    rdepend = bindb.aux_get("dev-libs/A-2", ["RDEPEND"])[0]
                    self.assertTrue("dev-libs/M" in rdepend)
                    self.assertTrue("dev-libs/M-moved" not in rdepend)

                    selected_set.load()
                    self.assertTrue("dev-libs/M" not in selected_set)
                    self.assertTrue("dev-libs/M-moved" in selected_set)

                finally:
                    playground.cleanup()
